/* src/bwshare/mod_bwshare.c   2008-1-31   Alan Kennington. */
/* $Id: mod_bwshare.c,v 0.1.0.118 2008/01/30 16:15:26 akenning Exp $ */
/*-----------------------------------------------------------------------------
Copyright (C) 2000, 2001, 2002, 2003, 2005, 2006, 2008, Alan Kennington.
You may distribute this software under the terms of Alan Kennington's
modified Artistic Licence, as specified in the accompanying LICENCE file.
-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------
DISCLAIMER.
The author of this software disclaims any express or implied guarantee of the
fitness of this software for any purpose. In no event shall the author of this
software be held liable for any direct, indirect, incidental, special,
exemplary, or consequential damages (including, but not limited to, procurement
of substitute services; loss of use, data, or profits; or business
interruption) however caused and on any theory of liability, whether in
contract, strict liability, or tort (including negligence or otherwise) arising
in any way out of the use of this software, even if advised of the possibility
of such damage.
-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------
bwshare version 0.2.1.
This is an Apache module for balancing and limiting client bandwidth demand.
The objectives are fair sharing and economization of bandwidth.
At present, this module performs only per-client-IP-address bandwidth
throttling by means of temporary blocking, but a future version might use
queueing to delay access requests to simulate a lower access bit-rate.
This software does not contain any code from other Apache modules or any
other software by other authors.
-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------
Struct definitions in this file:

bws_context_t::
bws_host_t::
bws_net_t::
host_sort_t::
host_sort2_t::
bws_hash_t::
bws_shm_t::
netlim_map::
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Functions in this file:

crc8init
crc8calc
crc8calc_z
crc8calc_rl
crc16init
crc16calc
crc16calc_z
crc16calc_rl
ipstring
parse_decimal
parse_subnet
parse_netlim_key

bws_context_ctor
bws_host_init
bws_net_init
bws_net_parse
bws_net_add
bws_net_add_subnet

sem_init
sem_remove
sem_grab
sem_free
sem_value

bws_init
host_sort_compare
bws_garbage_disposal
bws_newhost
host_rdns_set

compare_ip
compare_idle
compare_tx
compare_rdns
compare_files
compare_bytes
compare_key
hosts_sort2

shm_init
shm_remove

plain_cleanup_handler
child_cleanup_handler
bwshare_init
bwshare_create_server
bwshare_merge_server

bw_tx1debt_max
bw_tx1cred_rate
bw_tx2debt_max
bw_tx2cred_rate
bw_throttle_off
bw_subnet_limit

trace_internal
bwshare_trace
bwshare_info
bwshare_access_check
bwshare_logger
bwshare_init_handler
bwshare_register_hooks
-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------
Caution.
It should be noted that since this module uses semaphores, there exists the
theoretical potential that the whole server could get locked up by a software
error in this module. The code has been written very carefully to avoid this
possibility, but be warned that the potential still logically exists.
In particular, if you modify this software, deadlock conditions could be
introduced. Therefore do not modify this module unless you have a good
familiarity with System V semaphores.

As an example, reverse DNS lookups in the bwshare_logger function have been
kept out of a semaphore-critical section because this could hold up all other
child processes if DNS requests take a long time (which they very often do),
then all other child processes would be held up until all lookups are complete.
-----------------------------------------------------------------------------*/

/* Apache header files. */
#include "httpd.h"
#include "http_config.h"
/* Need http_core.h for REMOTE_DOUBLE_REV definition. */
#include "http_core.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_log.h"
/* #include <ap_config.h> */

/* Flag for version >= 2.0.x, where x = 13 or 16 or something. */
/* MODULE_MAGIC_NUMBER_MAJOR is in http_config.h. */
#if MODULE_MAGIC_NUMBER_MAJOR >= 20010224
#define BW_AP_VERSION_1 0
#define BW_AP_VERSION_2 1
#else
#define BW_AP_VERSION_1 1
#define BW_AP_VERSION_2 0
#endif

#if BW_AP_VERSION_1
#if defined(WIN32) || defined(BEOS) || defined(OS2) || defined(NETWARE)
#define WE_HAVE_UNIX 0
#else
#define WE_HAVE_UNIX 1
#endif
#endif /* BW_AP_VERSION_1 */

#if BW_AP_VERSION_2
#if defined(WIN32) || defined(BEOS) || defined(OS2) || defined(NETWARE)
#define WE_HAVE_UNIXD_H 0
#else
#define WE_HAVE_UNIXD_H 1
#endif
#endif /* BW_AP_VERSION_2 */

#if BW_AP_VERSION_1
/* File "http_conf_globals.h" contains ap_user_id, ap_group_id. */
#include "http_conf_globals.h"
#endif

#if BW_AP_VERSION_2
#include "apr_pools.h"
#endif

/* Apache 1.x.y doesn't have the unixd.h file (apparently). */
#if WE_HAVE_UNIXD_H
#include "unixd.h"
#endif

/* System header files. */
#include <sys/ipc.h>
#include <sys/types.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <ctype.h>
#include <stdlib.h>
#if WE_HAVE_UNIXD_H
#include <unistd.h>
#endif
/* #include <time.h> */

#if (defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)) \
 || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
/* union semun is defined by including <sys/sem.h> */
#else
/* Declaration for semctl() arg. Why not in header? What is posix anyway? */
union semun {
    int                 val;        /* value for SETVAL */
    struct semid_ds*    buf;        /* buffer for IPC_STAT, IPC_SET */
    unsigned short int* array;      /* array for GETALL, SETALL */
    struct seminfo*     __buf;      /* buffer for IPC_INFO */
    };
#endif

/*-----------------------------------------------------------------------------
Forward reference. The real structure is at the end of this file.
Entry-point table for this module. External linkage. Everything else is local.
-----------------------------------------------------------------------------*/
module bwshare_module;

/* Version of this module. */
#define BWSHARE_VERSION     "0.2.1"

/* Some basic numerical types. */
typedef unsigned char       uint8;
typedef unsigned short      uint16;
typedef unsigned long       uint32;
typedef unsigned long long  uint64;

typedef short               int16;
typedef long                int32;
typedef long long           int64;

/*-----------------------------------------------------------------------------
The context descriptor for servers and directories.
One of these is created for each server context and given to the Apache core.
The function ap_get_module_config() returns the current context when needed.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
This structure could record the parameters for the sharing of bandwidth for a
particular virtual host and/or directory/location.
For instance, there could be a flag to indicate whether any kind of limiting at
all is required for a given context, and if it is, then there could be
parameters like the maximum "TX debt" value that can be tolerated for that
context.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
There are two obvious approaches to using the location and vhost contexts.
1.  Have one client data table for _all_ locations/vhosts, and put some
    aggregate parameters into the context structure to place varying limits
    on behaviour.
2.  Have one client data table for _each_ location/vhost, and operate each
    location/vhost as if it were an independent physical host.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*    bws_context_t::   */
/*----------------------*/
typedef struct bws_context_t {
    int i;
    int throttle_off;       /* Set to non-zero to turn off throttling. */
    } bws_context_t;

/*-----------------------------------------------------------------------------
Some general statistical variables.

"n_init" counts the number of times that the server configures or reconfigures.
This is twice initially, and thereafter once for each "kill -HUP".

"n_server" is incremented each time a server context is notified.
This is done twice for each context initially, and then once for each
context again when the "kill -HUP" signal is sent.

"n_plain_cleanup" is supposed to be called when a pool is destroyed.

"n_child_cleanup" is supposed to be called when a pool is inherited by a child
process from another process, and it is about to delete the pool because an
"exec" is about to happen.

"n_handler" counts the number of times that the individual child process
handles requests. This is a different value for each child process.
-----------------------------------------------------------------------------*/
static uint32 n_shm_init = 0;
static uint32 n_sem_init = 0;
static uint32 n_init = 0;
static uint32 n_server = 0;
static uint32 n_plain_cleanup = 0;
static uint32 n_child_cleanup = 0;
static uint32 n_info = 0;
static uint32 n_trace = 0;
/* static uint32 n_logs = 0; */

/*-----------------------------------------------------------------------------
Semaphores.
If each virtual host is to be treated separately, it will be necessary to have a
separate remote-host history table and semaphore for each virtual host.
Each history table (type bws_shm_t) could be in a different shared-memory
segment or all in the same segment.
Either way, it would be better to have one semaphore per history table.
-----------------------------------------------------------------------------*/
/* For more than one semaphore, these variables must be restructured. */
static int sem_id = 0;              /* Handle for semaphore if created. */
static struct sembuf semop_grab;    /* Operation to grab all shared memory. */
static struct sembuf semop_free;    /* Operation to free all shared memory. */
static int sem_nesting_depth = 0;   /* Depth of nesting of semop_grab(). */

/*-----------------------------------------------------------------------------
Public keys could be used to implement inter-server bw-balancing.
They would also be useful for permitting monitoring by other processes.
In other words, there could be a stand-alone application which reads data
straight out of the shared memory so as to display the current status.
For instance, a CGI script/program could do this, invoked by the server.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set this macro to zero to use the magic numbers for IPC keys.
Do not set this macro to 0.
-----------------------------------------------------------------------------*/
#define IPC_KEYS_PRIVATE    1

#if ! IPC_KEYS_PRIVATE
/* This is just the string "bwsh". Probably won't be needed. */
#define SHM_KEY_BASE    0x62777368

/* This is just the string "BWSH". Probably won't be needed. */
#define SEM_KEY_BASE    0x42575348
#endif /* ! IPC_KEYS_PRIVATE */

/*-----------------------------------------------------------------------------
Shared memory data structures.
An appropriate number of hosts might be a mere few hundred, because on a
computer connected to a 33.6k modem (like mine), it is unlikely that more than
a hundred remote hosts will have histories that are of interest.
The table sizes should be made variable, depending on settings in the
httpd.conf file. This would imply that the host and hash tables should be
pointers rather than arrays.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If a remote host's TX debt falls to zero, it can be removed from the history
database.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Structure "bws_host_t" records the per-host current state.
-----------------------------------------------------------------------------*/
/* Set this to non-zero to force frequent garbage collections for testing. */
#define GARBAGE_TEST        0

#if ! GARBAGE_TEST
#define N_REMOTE_HOSTS      600
#else
#define N_REMOTE_HOSTS      32
#endif

#define NULL_HOST           (uint16)(-1)
#define N_DNS_CHARS         60      /* Size of reverse DNS cache string. */

/* Size of char-buffer for string formatting. */
#define STRBUFSIZE          100

/*---------------------------------------------------------------------------
Set this parameter to 1 to get r->bytes_sent throttling.
If BYTES_SENT == 0, throttling acts on request_rec::read_length.
If BYTES_SENT != 0, throttling acts on request_rec::bytes_sent.
It seems that "bytes_sent" is the number of bytes in the whole resource
which is requested, whereas "read_length" is the sum of the byte ranges
which are sent as actual content to the client.
Unfortunately, the read_length is always equal to zero, which I can't
understand from reading the Apache source. I do wish they would document
the API properly for us module writers!!!!!
Do not set this parameter to zero!
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
It seems that the read_length is the number of bytes read from the client
request, not the server response. So this is completely useless.
---------------------------------------------------------------------------*/
#define BYTES_SENT 1

/* High and low water marks. */
static const uint16 hwm_hard_deft = N_REMOTE_HOSTS - 1;
static const uint16 hwm_soft_deft = (3 * N_REMOTE_HOSTS) / 4;
static const uint16 lwm_hard_deft = ((3 * N_REMOTE_HOSTS) / 4) - 1;
static const uint16 lwm_soft_deft = N_REMOTE_HOSTS / 2;

#if BW_AP_VERSION_2
/* Set this to 1 to make bwshare_init get called from a new location. */
/* I'll explain this when the Apache2 developers document the module API. */
#define BW_INIT_NEW_LOCATION 1
#endif

/*---------------------------------------------------------------------------
A per-IP record which records the history of the request behaviour of a single
requesting IP address. This is all IPv4. Maybe in future there will be IPv6.
---------------------------------------------------------------------------*/
/*----------------------*/
/*      bws_host_t::    */
/*----------------------*/
typedef struct bws_host_t {
    uint16              next;         /* Array index of next record. */
    uint16              hash_index;   /* Hash index for containing list. */

    uint32              address;      /* IP address of client. */
    uint32              n_files;      /* Total file requests. */
    uint32              n_bytes;      /* Total transmitted bytes. */
    uint32              last_time;    /* Last update time. Secs since 1970. */

    double              tx1debt;      /* Current file debt of this host. */
    double              tx1debtpeak;  /* Max current file debt of this host. */
    double              tx2debt;      /* Current byte debt of this host. */
    double              tx2debtpeak;  /* Max current byte debt of this host. */

    /* BW balancing/limiting parameters. */
    double              tx1cred_rate; /* Files credit per second. */
    double              tx1debt_max;  /* Max permissible debt. */
    double              tx2cred_rate; /* Bytes credit per second. */
    double              tx2debt_max;  /* Max permissible debt. */

    char                rdns[N_DNS_CHARS]; /* Reverse DNS translation. */
    } bws_host_t;

/*---------------------------------------------------------------------------
This structure represents an IPv4 subnet of the Internet.
Any old 32-bit bit-mask may be used, not just left-justified masks.
An IP address x is in the subnet if and only if

    x & mask  ==  address & mask

This is the same as:  (x ^ address) & mask == 0.
The number of addresses in the subnet is the number of bits in "mask".
The case "mask" == 0xffffffff contains the single host: "address".

If the address/mask fits the IP address x of the HTTP client, the
BW parameters are applied to that address.
The first network which matches the IP address is used.
Therefore order is important.
---------------------------------------------------------------------------*/
/*----------------------*/
/*      bws_net_t::     */
/*----------------------*/
typedef struct bws_net_t {
    uint32              address;      /* Address of IP subnet. */
    uint32              mask;         /* Mask for IP subnet. */

    /* BW balancing/limiting parameters. */
    double              tx1cred_rate; /* Files credit per second. */
    double              tx1debt_max;  /* Max permissible debt. */
    double              tx2cred_rate; /* Bytes credit per second. */
    double              tx2debt_max;  /* Max permissible debt. */
    } bws_net_t;

/* If the web server really needs more than 500 subnet, increase this number. */
#define N_NETWORKS          500

/*-----------------------------------------------------------------------------
A structure for sorting elements of the host table in preparation for
garbage removal.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*     host_sort_t::    */
/*----------------------*/
typedef struct host_sort_t {
    uint16              host_index; /* Array index of this host record. */
    uint16              keep;       /* Indication of whether to keep it. */
    int32               value;      /* How valuable the record is to keep. */
    } host_sort_t;

/*-----------------------------------------------------------------------------
A structure for sorting elements of the host table for display.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*    host_sort2_t::    */
/*----------------------*/
typedef struct host_sort2_t {
    double              tx_debt;    /* Some sort of debt. */
    bws_host_t*         host;       /* The host record. */
    uint16              host_index; /* Array index of this host record. */
    } host_sort2_t;

#define N_HASH_ENTRIES      2000
#define NULL_HASH           (uint16)(-1)

/*----------------------*/
/*      bws_hash_t::    */
/*----------------------*/
typedef struct bws_hash_t {
    uint16              head;       /* Head of list for given hash key. */
                                    /* No padding in this structure. */
    /* Surely I can think of something else to put here! */
    } bws_hash_t;

/*-----------------------------------------------------------------------------
A single data structure to be shared among all child processes.
A single copy of this is located in the shared memory segment.
This is a kind of common blackboard which all the processes read and write to.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Probably should use integer indices rather than pointers in the host record
heap.
-----------------------------------------------------------------------------*/
#if N_REMOTE_HOSTS > 65535
#error N_REMOTE_HOSTS must not exceed 65535 = 2^16 - 1.
#endif
#if N_REMOTE_HOSTS < 1
#error N_REMOTE_HOSTS must be at least 1.
#endif

#if N_NETWORKS > 65535
#error N_NETWORKS must not exceed 65535 = 2^16 - 1.
#endif
#if N_NETWORKS < 1
#error N_NETWORKS must be at least 1.
#endif

#if N_HASH_ENTRIES > 65535
#error N_HASH_ENTRIES must not exceed 65535 = 2^16 - 1.
#endif
#if N_HASH_ENTRIES < 1
#error N_HASH_ENTRIES must be at least 1.
#endif

/*----------------------*/
/*      bws_shm_t::     */
/*----------------------*/
typedef struct bws_shm_t {
    time_t              start_time;             /* Start of hosts database. */
    bws_host_t          hosts[N_REMOTE_HOSTS];  /* Host record heap. */
    uint16              hosts_free;             /* First of free list. */

    uint16              hwm_hard;               /* Hard high-water mark. */
    uint16              hwm_soft;               /* Soft high-water mark. */
    uint16              lwm_hard;               /* Hard low-water mark. */
    uint16              lwm_soft;               /* Soft low-water mark. */

    /* A hash table of non-free-list head pointers. Hopefully 2 bytes each. */
    bws_hash_t          hashtable[N_HASH_ENTRIES];

    /* General stats. */
    uint32              n_info;                 /* Number of info calls. */
    uint32              n_trace;                /* Number of trace calls. */
    uint32              n_logs;                 /* Number of logger calls. */
    uint64              n_bytes;                /* Number of bytes served. */

    /* BW balancing/limiting parameters. Global default. */
    double              tx1cred_rate;           /* Files credit per second. */
    double              tx1debt_max;            /* Max permissible debt. */
    double              tx2cred_rate;           /* Bytes credit per second. */
    double              tx2debt_max;            /* Max permissible debt. */

    uint32              n_files_warn;           /* Criterion for warning. */
    uint32              n_bytes_warn;           /* Criterion for warning. */
    uint32              idle_show;              /* Recent time definition. */

    /* IP networks which get special treatment. */
    bws_net_t           networks[N_NETWORKS];
    uint16              n_networks;
    } bws_shm_t;

/* Attribute to sort "hosts" table by. */
typedef enum host_sort_enum {
    sort_ip,            /* Client IP address. */
    sort_idle,          /* Idle time of client. */
    sort_tx1,           /* TX 1 debt. */
    sort_tx2,           /* TX 2 debt. */
    sort_tx1peak,       /* Peak TX 1 debt. */
    sort_tx2peak,       /* Peak TX 2 debt. */
    sort_rdns,          /* Reverse DNS translation. (Not yet implemented.) */
    sort_files,         /* Number of files. */
    sort_bytes,         /* Number of bytes. */
    sort_key,           /* Hash key. */
    sort_index,         /* Index in hosts array. */
    sort_MAX            /* Number of sort types. */
    } host_sort_enum;

/* Keywords in the BW_subnet_limit configuration parameter. */
typedef enum netlim_enum {
    netlim_net,         /* net, network, host, hosts. */
    netlim_tx1rate,     /* tx1rate. */
    netlim_tx1max,      /* tx1max. */
    netlim_tx2rate,     /* tx2rate. */
    netlim_tx2max,      /* tx2max. */

    netlim_bit,         /* bit, bits. */
    netlim_byte,        /* byte, bytes. */
    netlim_file,        /* file, files, request, requests. */

    netlim_sec,         /* sec, secs, second, seconds. */
    netlim_min,         /* min, mins, minute, minutes. */
    netlim_hr,          /* hr, hrs, hour, hours. */
    netlim_day,         /* day, days. */

    netlim_MAX          /* Number of netlimit types. */
    } netlim_enum;

/* Table for parsing netlimit keywords. */
/*----------------------*/
/*     netlim_map::     */
/*----------------------*/
typedef struct netlim_map {
    netlim_enum         key;
    char*               name;
    } netlim_map;

/* Look-up table for BW_subnet_limit keywords. Who needs hash tables!!? */
/* Must terminate with a null string and key == netlim_MAX. */
static netlim_map netlim_maps[] = {
    { netlim_net,       "net" },
    { netlim_net,       "network" },
    { netlim_net,       "host" },
    { netlim_net,       "hosts" },

    { netlim_tx1rate,   "tx1rate" },
    { netlim_tx1max,    "tx1max" },
    { netlim_tx2rate,   "tx2rate" },
    { netlim_tx2max,    "tx2max" },

    { netlim_bit,       "bit" },
    { netlim_bit,       "bits" },

    { netlim_byte,      "byte" },
    { netlim_byte,      "bytes" },

    { netlim_file,      "file" },
    { netlim_file,      "files" },
    { netlim_file,      "request" },
    { netlim_file,      "requests" },

    { netlim_sec,       "sec" },
    { netlim_sec,       "secs" },
    { netlim_sec,       "second" },
    { netlim_sec,       "seconds" },

    { netlim_min,       "min" },
    { netlim_min,       "mins" },
    { netlim_min,       "minute" },
    { netlim_min,       "minutes" },

    { netlim_hr,        "hr" },
    { netlim_hr,        "hrs" },
    { netlim_hr,        "hour" },
    { netlim_hr,        "hours" },

    { netlim_day,       "day" },
    { netlim_day,       "days" },

    { netlim_MAX,       (char*)0 }
    };
#define N_NETLIM_MAPS (sizeof(netlim_maps)/sizeof(netlim_map))

/*-----------------------------------------------------------------------------
Local "global variables" for shared memory.
Note that in the case of multiple remote-host history contexts (type
bws_shm_t), these could all be placed in a single shared memory segment by
defining "shm_base" to be an array of history contexts rather than a single
context.
-----------------------------------------------------------------------------*/
static int shm_id = 0;
static bws_shm_t* shm_base = 0;

/* Default BW balancing/limiting parameters. Using fixed-point arithmetic. */
static const double tx1cred_rate_deft = 0.1333; /* Files per second. */
static const double tx1debt_max_deft = 20;      /* Max requests debt. */

static const double tx2cred_rate_deft = 2500;   /* Bytes per second. */
static const double tx2debt_max_deft = 3000000; /* Max bytes debt. */

/* These should be made settable in apache configurations files. */
static const uint32 n_files_warn_deft = 50;     /* Criterion for warning. */
static const uint32 n_bytes_warn_deft = 1000000; /* Criterion for warning. */
static const uint32 idle_show_deft = 600;       /* Criterion for display. */

static const host_sort_enum hosts_sort = sort_idle; /* Hosts sort field. */

static int show_key = 0;                    /* Show column of keys. */
static int show_index = 0;                  /* Show column of index values. */
static int show_key_total = 0;              /* Show total of keys. */
static int show_index_total = 0;            /* Show total of index values. */

/* Display parameters. The should also settable in apache config files. */
#define C_SKYBLUE           "#89beff"       /* Sky blue. */
#define C_PEPPERMINT        "#a8ffd0"       /* Bright peppermint green. */
#define C_MSBLUE            "#d1e1ff"       /* Light sky blue. */
#define C_LGREEN            "#e5d995"       /* Light green. */
#define C_LDIRT             "#e8d196"       /* Light dirt. */
#define C_DGOLD             "#efbf3b"       /* Darkish gold. */
#define C_BEIGE             "#f2c87b"       /* Beige. */
#define C_LEMON             "#fcff59"       /* Lemon-yellow. */
#define C_LBROWN            "#fee4b3"       /* Cafe latte. */
#define C_LRED              "#ff615e"       /* Slightly faded red. */
#define C_DYELLOW           "#ffe16d"       /* Dark yellow. */
#define C_YBEIGE            "#ffe798"       /* Yellowish beige. */
#define C_DARKGREY          "#333333"       /* Dark grey. */

#define COLOUR_INFO_BG      C_BEIGE         /* Whole-page background. */
#define COLOUR_INFO_TAB0    C_LDIRT         /* Parameter summary background. */
#define COLOUR_INFO_FRM_BD  C_SKYBLUE       /* Border around form. */
#define COLOUR_INFO_FRM_BG  C_DYELLOW       /* Form background. */
#define COLOUR_INFO_FRM_FADE C_DARKGREY     /* Fade out undefined option. */

#if 0
#define COLOUR_TRACE_BG     C_PEPPERMINT    /* Whole-page background. */
#endif
#define COLOUR_TRACE_BG     C_LGREEN        /* Whole-page background. */
#define COLOUR_TRACE_TAB0   C_LDIRT         /* Parameter summary background. */
#define COLOUR_TRACE_TAB1   C_YBEIGE        /* Host table background. */
#define COLOUR_TRACE_BAD    C_LRED          /* Excessive debt. */
#define COLOUR_TRACE_WARN   C_LEMON         /* Debt warning. */

#if BW_AP_VERSION_2
#define BW_CONN_TRACE 0
#if BW_CONN_TRACE
/*-------------------------------------------------------------------------
Some ad hoc data for debugging the Apache 2.0.16 API.
-------------------------------------------------------------------------*/
/*----------------------*/
/*     conn_trace_t::   */
/*----------------------*/
typedef struct conn_trace_t {
    sa_family_t         family;
    apr_port_t          port;
    apr_socklen_t       salen;
    int                 ipaddr_len;
    int                 addr_str_len;
    } conn_trace_t;

/* Info from last invocations of two functions. */
static conn_trace_t ctt_handler;
static conn_trace_t ctt_logger;
#endif
#endif /* BW_AP_VERSION_2 */

/*-------------------------------------------------------------------------
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
CRC (Cyclic Redundancy Code) functions for use in hash algorithms.
Note that most of these CRC functions are not used.
Only the crc16calc_rl() function is used at present.
The other stuff might come in useful for later development though.
-----------------------------------------------------------------------------*/
/* CRC algorithm constants. */
static const int crc8poly = 0x107;          /* As in ATM HCS field. */
static const uint16 crc16poly = 0x1021;     /* Similar to PPP poly. */
static const uint16 crc16begin = 0xffff;    /* Initial value of register. */

#if 0
/*-----------------------------------------------------------------------------
Lookup table for CRC-8 (poly = x^8 + x^2 + x^1 + x^0 = 0x107):
(crc8table[x] = remainder of x * 2^8 divided by 0x107.)
-----------------------------------------------------------------------------*/
static uint8 crc8table[256] = {
    0x00, 0x07, 0x0e, 0x09, 0x1c, 0x1b, 0x12, 0x15, 0x38, 0x3f, 0x36, 0x31,
    0x24, 0x23, 0x2a, 0x2d, 0x70, 0x77, 0x7e, 0x79, 0x6c, 0x6b, 0x62, 0x65,
    0x48, 0x4f, 0x46, 0x41, 0x54, 0x53, 0x5a, 0x5d, 0xe0, 0xe7, 0xee, 0xe9,
    0xfc, 0xfb, 0xf2, 0xf5, 0xd8, 0xdf, 0xd6, 0xd1, 0xc4, 0xc3, 0xca, 0xcd,
    0x90, 0x97, 0x9e, 0x99, 0x8c, 0x8b, 0x82, 0x85, 0xa8, 0xaf, 0xa6, 0xa1,
    0xb4, 0xb3, 0xba, 0xbd, 0xc7, 0xc0, 0xc9, 0xce, 0xdb, 0xdc, 0xd5, 0xd2,
    0xff, 0xf8, 0xf1, 0xf6, 0xe3, 0xe4, 0xed, 0xea, 0xb7, 0xb0, 0xb9, 0xbe,
    0xab, 0xac, 0xa5, 0xa2, 0x8f, 0x88, 0x81, 0x86, 0x93, 0x94, 0x9d, 0x9a,
    0x27, 0x20, 0x29, 0x2e, 0x3b, 0x3c, 0x35, 0x32, 0x1f, 0x18, 0x11, 0x16,
    0x03, 0x04, 0x0d, 0x0a, 0x57, 0x50, 0x59, 0x5e, 0x4b, 0x4c, 0x45, 0x42,
    0x6f, 0x68, 0x61, 0x66, 0x73, 0x74, 0x7d, 0x7a, 0x89, 0x8e, 0x87, 0x80,
    0x95, 0x92, 0x9b, 0x9c, 0xb1, 0xb6, 0xbf, 0xb8, 0xad, 0xaa, 0xa3, 0xa4,
    0xf9, 0xfe, 0xf7, 0xf0, 0xe5, 0xe2, 0xeb, 0xec, 0xc1, 0xc6, 0xcf, 0xc8,
    0xdd, 0xda, 0xd3, 0xd4, 0x69, 0x6e, 0x67, 0x60, 0x75, 0x72, 0x7b, 0x7c,
    0x51, 0x56, 0x5f, 0x58, 0x4d, 0x4a, 0x43, 0x44, 0x19, 0x1e, 0x17, 0x10,
    0x05, 0x02, 0x0b, 0x0c, 0x21, 0x26, 0x2f, 0x28, 0x3d, 0x3a, 0x33, 0x34,
    0x4e, 0x49, 0x40, 0x47, 0x52, 0x55, 0x5c, 0x5b, 0x76, 0x71, 0x78, 0x7f,
    0x6a, 0x6d, 0x64, 0x63, 0x3e, 0x39, 0x30, 0x37, 0x22, 0x25, 0x2c, 0x2b,
    0x06, 0x01, 0x08, 0x0f, 0x1a, 0x1d, 0x14, 0x13, 0xae, 0xa9, 0xa0, 0xa7,
    0xb2, 0xb5, 0xbc, 0xbb, 0x96, 0x91, 0x98, 0x9f, 0x8a, 0x8d, 0x84, 0x83,
    0xde, 0xd9, 0xd0, 0xd7, 0xc2, 0xc5, 0xcc, 0xcb, 0xe6, 0xe1, 0xe8, 0xef,
    0xfa, 0xfd, 0xf4, 0xf3
    }; /* End of array crc8table. */
#endif /* 0 */

/*-----------------------------------------------------------------------------
Lookup table for CRC-16 (poly = 0x1 1021):
(crc16table[x] = remainder of x * 2^16 divided by  0x11021, with
the log division register initialised to 0xffff.)
-----------------------------------------------------------------------------*/
static uint16 crc16table[256] = {
    0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
    0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
    0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
    0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
    0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
    0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
    0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
    0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
    0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
    0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
    0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
    0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
    0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
    0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
    0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
    0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
    0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
    0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
    0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
    0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
    0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
    0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
    0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
    0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
    0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
    0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
    0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
    0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
    0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
    0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
    0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
    0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0
    }; /* End of array crc16table. */

#if 0
/*-----------------------------------------------------------------------------
crc8init() (re-)initialises the crc8table. This can be called to verify that
there are no typing errors in the table.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*       crc8init       */
/*----------------------*/
static int crc8init(void) {
    int i_error_first = -1;
    int i = 0;

    for (i = 0; i < 256; ++i) {
        unsigned int r = i << 8;        /* r is the remainder. */
        int j = 0;

        /* The following loop is just long division: */
        for (j = 7; j >= 0; --j)
            if (r & (1 << (j + 8)))
                r ^= crc8poly << j;

        /* Check for error in the previous value: */
        if ((crc8table[i] != (r & 0xff)) && i_error_first < 0)
            i_error_first = i;
        crc8table[i] = r & 0xff;
        }
    if (i_error_first >= 0) {
#if 0
        fprintf(stderr, "Found error in crc8table[%d].\n", i_error_first);
#endif
        return -1;
        }
    return 0;
    } /* End of function crc8init. */

/*-----------------------------------------------------------------------------
crc8calc() returns the" remainder" of the given byte array "divided" by the
"polynomial" 0x107, and XORed with 0x55.
The given array has address "buf" and length "len".
This CRC is the same as used in (the 5th byte of) ATM cell headers.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*       crc8calc       */
/*----------------------*/
static char crc8calc(const char* buf, int len) {
    char r = 0;             /* r is the remainder. */
    if (buf && len > 0)
        while (--len >= 0)  /* Do long division, 8 bits at a time. */
            r = crc8table[(r ^ *buf++) & 0xff];
    return r ^ 0x55;
    } /* End of function crc8calc. */

/*-----------------------------------------------------------------------------
crc8calc() returns the" remainder" of the given byte array "divided" by the
"polynomial" 0x107, and XORed with 0x55.
The given array has address "buf" and is null-terminated.
This CRC is the same as used in (the 5th byte of) ATM cell headers.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*      crc8calc_z      */
/*----------------------*/
static char crc8calc_z(const char* buf) {
    char r = 0;             /* r is the remainder. */
    if (buf)
        while (*buf)  /* Do long division, 8 bits at a time. */
            r = crc8table[(r ^ *buf++) & 0xff];
    return r ^ 0x55;
    } /* End of function crc8calc_z. */

/*-----------------------------------------------------------------------------
This does a right-to-left CRC-8 on the given 32-bit integer.
This is suitable for use as a hash function from 32 to 8 bits.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*     crc8calc_rl      */
/*----------------------*/
static uint8 crc8calc_rl(uint32 data) {
    char r = crc8table[data & 0xff];
    data >>= 8;
    r = crc8table[(r ^ data) & 0xff];
    data >>= 8;
    r = crc8table[(r ^ data) & 0xff];
    data >>= 8;
    r = crc8table[(r ^ data) & 0xff];
    return (uint8)(r ^ 0x55);
    } /* End of function crc8calc_rl. */

/*-----------------------------------------------------------------------------
crc16init() (re-)initialises the crc16table. This can be called to verify that
there are no typing errors in the table.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*       crc16init      */
/*----------------------*/
static int crc16init(uint16 poly, int trace) {
    int i_error_first = -1;
    int i = 0;
    for (i = 0; i < 256; ++i) {
        /* Initialize the remainder: */
        uint16 r = i << 8;
        int j = 0;

        /* The following loop is just long division: */
        for (j = 7; j >= 0; --j)
            if (r & 0x8000) {
                r <<= 1;
                r ^= poly;
                }
            else
                r <<= 1;

        /* Check for error in the previous value: */
        if ((crc16table[i] != r) && i_error_first < 0)
            i_error_first = i;
        crc16table[i] = r;
        }
    if (i_error_first >= 0) {
/*        cout << "Found error in crc16table[" << i_error_first << endl; */
        return -1;
        }
    return 0;
    } /* End of function crc16init. */

/*-----------------------------------------------------------------------------
crc16calc() returns the" remainder" of the given byte array "divided" by the
"polynomial" crc16poly, modulo some fine points.
The given array has address "buf" and length "len".
This CRC is the same as used in PPP frames. But in PPP it may or may not be
sent in reverse bit order, depending on one's perspective.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*       crc16calc      */
/*----------------------*/
static uint16 crc16calc(const char* buf, int len) {
    register uint16 crc = crc16begin;
    if (buf)
        while (len > 0) { /* Do long division, 8 bits at a time. */
            crc = (crc << 8) ^ crc16table[((crc >> 8) ^ *buf++) & 0xff];
            len -= 1;
            }
    return ~crc & 0xffff;
    } /* End of function crc16calc. */

/*-----------------------------------------------------------------------------
crc16calc() returns the" remainder" of the given byte array "divided" by the
"polynomial" crc16poly, modulo some fine points.
The given array has address "buf" and is (8-bit) null-terminated.
This CRC is the same as used in PPP frames. But in PPP it may or may not be
sent in reverse bit order, depending on one's perspective.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*      crc16calc_z     */
/*----------------------*/
static uint16 crc16calc_z(const char* buf) {
    register uint16 crc = crc16begin;
    if (buf)
        while (*buf)    /* Do long division, 8 bits at a time. */
            crc = (crc << 8) ^ crc16table[((crc >> 8) ^ *buf++) & 0xff];
    return ~crc & 0xffff;
    } /* End of function crc16calc_z. */
#endif /* 0 */

/*-----------------------------------------------------------------------------
This does a right-to-left CRC-16 on the given 32-bit integer.
This is suitable for use as a hash function from 32 to 16 bits.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*     crc16calc_rl     */
/*----------------------*/
static uint16 crc16calc_rl(uint32 data) {
    register uint16 crc = crc16begin;
    crc = (crc << 8) ^ crc16table[((crc >> 8) ^ (data & 0xff)) & 0xff];
    data >>= 8;
    crc = (crc << 8) ^ crc16table[((crc >> 8) ^ (data & 0xff)) & 0xff];
    data >>= 8;
    crc = (crc << 8) ^ crc16table[((crc >> 8) ^ (data & 0xff)) & 0xff];
    data >>= 8;
    crc = (crc << 8) ^ crc16table[((crc >> 8) ^ (data & 0xff)) & 0xff];

    return ~crc & 0xffff;
    } /* End of function crc16calc_rl. */

/*-----------------------------------------------------------------------------
This function returns static memory. So don't try to free it or over-write it.
Not thread-safe.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*       ipstring       */
/*----------------------*/
static const char* ipstring(unsigned long x) {
    static char buf[16] = "...............";

    sprintf(buf, "%lu.%lu.%lu.%lu",
        (x >> 24) & 0xff, (x >> 16) & 0xff, (x >> 8) & 0xff, x & 0xff);
    return buf;
    } /* End of function ipstring. */

/*-----------------------------------------------------------------------------
Convert a uint64 to a comma-separated string.
This function returns static memory. So don't try to free it or over-write it.
Not thread-safe.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*  uint64comma_string  */
/*----------------------*/
static char* uint64comma_string(unsigned long x) {
    static char strbuf[STRBUFSIZE] = ""; /* Largish string buffer. */
    unsigned long i = x;
    char* pc = strbuf + STRBUFSIZE;
    int ndig = 0;

    /* 2^64 - 1 = 18,446,744,073,709,551,615. */
    *--pc = 0;
    for (;;) {
        *--pc = '0' + (i % 10);
        i /= 10;
        if (i <= 0)
            break;
        ndig += 1;
        if (i > 0 && (ndig % 3) == 0)
            *--pc = ',';
        }
    return pc;
    } /* End of function uint64comma_string. */

/*---------------------------------------------------------------------------
Parse a non-negative decimal integer.
Return pointer to after where it's parsed.
In case of failure, return str1.
NOTE: If the integer is really big, it will overflow and give the wrong answer!
---------------------------------------------------------------------------*/
/*----------------------*/
/*     parse_decimal    */
/*----------------------*/
static const char* parse_decimal(const char* str1, unsigned long* result) {
    const char* str3;
    unsigned long n = 0;
    int bad = 0;

    if (!str1 || !result)
        return str1;
    str3 = str1;

    /* Get the decimal number. */
    while (*str3 && isdigit(*str3)) {
        unsigned long digit = *str3 - '0';  /* Assume ascii or similar. */
        if (n > LONG_MAX/10) {
            bad = 1;
            break;
            }
        n *= 10;
        if (n > LONG_MAX - digit) {
            bad = 1;
            break;
            }
        n += digit;
        ++str3;
        }
    if (bad)
        return str1;

    *result = n;
    return str3;
    } /* End of function parse_decimal. */

/*---------------------------------------------------------------------------
Parse an IPv4 subnet.
Expect something like 58.6.65.66/28 or 58.6.65.66/0xfffffff0.
If something goes wrong, str1 is returned and address/mask are unchanged.
Otherwise the return value points to the character after the last parsed
character and the address and mask are both set.
If there is no slash postfix, the mask is set to 0xffffffff.
---------------------------------------------------------------------------*/
/*----------------------*/
/*     parse_subnet     */
/*----------------------*/
static const char* parse_subnet(const char* str1,
        uint32* p_address, uint32* p_mask) {
    uint32 address = 0;
    uint32 mask = 0;
    int j = 0;
    int bad = 0;
    int nslashes = 0;
    unsigned long n = 0;
    const char* str5 = 0;
    const char* str0 = 0;

    if (!str1 || !p_address || !p_mask)
        return str1;
    str0 = str1;

    /* Skip white space. */
    while (*str0 && isspace(*str0))
        ++str0;
    if (!*str0)
        return str1;

    /* First must see a dotted-quad-decimal address. */
    for (j = 0; j < 4; ++j) {
        const char* str3;
        int ndots;

        /* Get the decimal number. */
        n = 0;
        str3 = parse_decimal(str0, &n);
        if (str3 == str0 || n > 255) {
            bad = 1;
            break;
            }
        str0 = str3;

        /* Shift the integer into the address. */
        address <<= 8;
        address |= (n & 0xff);

        if (j >= 3)
            break;

        /* Skip the decimal point. */
        ndots = 0;
        while (*str0 && (isspace(*str0) || *str0 == '.')) {
            if (*str0 == '.')
                ndots += 1;
            ++str0;
            }
        /* End of args or wrong number of dots. */
        if (!*str0 || ndots != 1) {
            bad = 1;
            break;
            }
        }
    if (bad)
        return str1;

    /* Now we expect an optional mask. */
    nslashes = 0;
    while (*str0 && (isspace(*str0) || *str0 == '/')) {
        if (*str0 == '/')
            nslashes += 1;
        ++str0;
        }
    /* If there's no slash, it's a single IP address. */
    if (nslashes <= 0 || !*str0) {
        *p_address = address;
        *p_mask = 0xffffffff;
        return str0;
        }
    /* Greater than one slash is an error. */
    if (nslashes > 1)
        return str1;

    /* After a slash, expect decimal integer or hexadecimal mask. */
    if (*str0 == '0' && *(str0+1) && tolower(*(str0+1)) == 'x') {
        str0 += 2;

        /* Maybe we have a hexadecimal bit mask. */
        mask = 0;
        for (j = 0; j < 8; ++j) {
            if (*str0 && isxdigit(*str0)) {
                int digit = 0;
                /* Assume ascii or similar. */
                if ('0' <= *str0 && *str0 <= '9')
                    digit = *str0 - '0';
                else if ('a' <= *str0 && *str0 <= 'f')
                    digit = *str0 - 'a';
                else if ('A' <= *str0 && *str0 <= 'F')
                    digit = *str0 - 'A';
                mask <<= 4;
                mask |= (digit & 0x0f);
                ++str0;
                }
            else
                break;
            }
        /* Must get the full 8 hex digits. */
        if (j < 8)
            return str1;

        /* If there's another hex digit after the 8 digits, that's an error. */
        if (*str0 && isxdigit(*str0))
            return str1;

        /* Otherwise we're home and hosed. */
        *p_address = address;
        *p_mask = mask;
        return str0;
        }

    /* Now it must be an integer in the interval [0,32]. */
    str5 = parse_decimal(str0, &n);
    if (str5 == str0 || n > 32)
        return str1;
    str0 = str5;
    mask = 0xffffffff << (32 - n);

    *p_address = address;
    *p_mask = mask;
    return str0;
    } /* End of function parse_subnet. */

/*---------------------------------------------------------------------------
Parse a keyword.
If successful, *key is set to the result, and the address after the parsed
keyword is returned. Otherwise str1 is returned and *key is unchanged.
---------------------------------------------------------------------------*/
/*----------------------*/
/*   parse_netlim_key   */
/*----------------------*/
static const char* parse_netlim_key(const char* str1, netlim_enum* pkey) {
    const char* str0 = 0;
    const char* str2 = 0;
    int len2 = 0;
    int found = 0;
    int ii;

    if (!str1 || !pkey)
        return str1;
    str0 = str1;

    /* Skip white space. */
    while (*str0 && isspace(*str0))
        ++str0;
    /* Exit at end of string. */
    if (!*str0)
        return str1;

    /* Examine the first token. Put str2 just after the token. */
    str2 = str0;
    while (*str2 && (isalnum(*str2) || *str2 == '_'))
        ++str2;
    len2 = str2 - str0;

    /* If there's no alphabetic identifier, it's an error. */
    if (len2 <= 0)
        return str1;

    /* Look up the identifier in the table. */
    found = 0;
    for (ii = 0; ii < N_NETLIM_MAPS; ++ii) {
        netlim_map* p = &netlim_maps[ii];
        if (p->key == netlim_MAX || !p->name)
            break;
        if (strncasecmp(str0, p->name, len2) == 0) {
            found = 1;
            break;
            }
        }
    /* If keyword is not recognized, give up. */
    if (!found)
        return str1;

    *pkey = netlim_maps[ii].key;
    return str2;
    } /* End of function parse_netlim_key. */

/*---------------------------------------------------------------------------
Constructor for object of type bws_context_t.
---------------------------------------------------------------------------*/
/*----------------------*/
/*   bws_context_ctor   */
/*----------------------*/
static int bws_context_ctor(bws_context_t* obj1) {
    if (! obj1)
        return -1;

    /* Number of servers??? This is never used. It's a place-holder. */
    obj1->i = 0;
    obj1->throttle_off = 0;

    return 0;
    } /* End of function bws_context_ctor. */

/*---------------------------------------------------------------------------
Initializer for object of type bws_host_t.
---------------------------------------------------------------------------*/
/*----------------------*/
/*     bws_host_init    */
/*----------------------*/
static int bws_host_init(bws_host_t* obj1) {
    if (! obj1)
        return -1;

    obj1->next = NULL_HOST;
    obj1->hash_index = NULL_HASH;   /* Must have less than 65536 hashes. */

    obj1->address = 0;
    obj1->n_files = 0;
    obj1->n_bytes = 0;
    obj1->last_time = 0;

    obj1->tx1debt = 0;
    obj1->tx1debtpeak = 0;
    obj1->tx2debt = 0;
    obj1->tx2debtpeak = 0;

    obj1->tx1cred_rate = 0;
    obj1->tx1debt_max = 0;
    obj1->tx2cred_rate = 0;
    obj1->tx2debt_max = 0;

    memset(obj1->rdns, 0, N_DNS_CHARS);   /* Clear the DNS cache area. */
    return 0;
    } /* End of function bws_host_init. */

/*---------------------------------------------------------------------------
Initializer for object of type bws_net_t.
---------------------------------------------------------------------------*/
/*----------------------*/
/*     bws_net_init     */
/*----------------------*/
static int bws_net_init(bws_net_t* obj1) {
    if (! obj1)
        return -1;

    obj1->address = 0;
    obj1->mask = 0;

    obj1->tx1cred_rate = 0;
    obj1->tx1debt_max = 0;
    obj1->tx2cred_rate = 0;
    obj1->tx2debt_max = 0;

    return 0;
    } /* End of function bws_net_init. */

/*---------------------------------------------------------------------------
Parse bandwidth limitation parameters from a configuration string.
Example:

    BW_subnet_limit net = 58.6.65.66/28 \
    tx1rate = 5.40 files/min  tx1max = 20.0 files \
    tx2rate = 2400 bytes/sec  tx2max = 3000000 bytes

Before calling this function, set all parameters to defaults.
---------------------------------------------------------------------------*/
/*----------------------*/
/*     bws_net_parse    */
/*----------------------*/
static int bws_net_parse(bws_net_t* obj1, const char* args) {
    const char* str1 = 0;
#if 0
#define NET_PARSE_BUFSIZE   128
    char buf[NET_PARSE_BUFSIZE];
#endif
    int retval = 0;

    if (!obj1)
        return -1;
    if (!args)
        return -2;

    /* Go round the parse loop once for each variable-setting. */
    str1 = args;
    while (*str1) {
        netlim_enum key0 = netlim_MAX;
        const char* str2 = 0;
        int bad = 0;
        int n_equals = 0;

        str2 = parse_netlim_key(str1, &key0);
        if (str2 == str1) {
            bad = 1;
            break;
            }
        str1 = str2;

        /* Skip an optional '=' and white space. */
        n_equals = 0;
        while (*str1 && (isspace(*str1) || *str1 == '=')) {
            if (*str1 == '=')
                n_equals += 1;
            ++str1;
            }
        /* Double equals sign is not accepted. */
        if (n_equals > 1) {
            retval = -5;
            break;
            }
        /* Exit at early end of string. */
        if (!*str1)
            break;

        /* Parse the arguments. */
        bad = 0;
        switch (key0) {
        case netlim_net:
            str2 = parse_subnet(str1, &obj1->address, &obj1->mask);
            if (!str2 || str2 == str1)
                bad = 1;
            else
                str1 = str2;
            break;
        case netlim_tx1rate:
        case netlim_tx2rate:
            {
            /* Expect something like "5.40 files/min" or "32000 bits/sec". */
            netlim_enum key1 = netlim_MAX;
            netlim_enum key2 = netlim_MAX;
            const char* str3 = 0;
            char* str4 = 0;
            int n_slashes = 0;

            double val = strtod(str1, &str4);
            if (!str4 || str4 == str1 || errno == ERANGE) {
                bad = 1;
                break;
                }
            str1 = str4;

            /* Fetch units of the rate, e.g. "files/min". */
            str3 = parse_netlim_key(str1, &key1);
            if (!str3 || str3 == str1) {
                bad = 1;
                break;
                }
            str1 = str3;

            /* Skip a mandatory '/' and white space. */
            n_slashes = 0;
            while (*str1 && (isspace(*str1) || *str1 == '/')) {
                if (*str1 == '/')
                    n_slashes += 1;
                ++str1;
                }
            if (n_slashes != 1) {
                bad = 1;
                break;
                }

            /* Parse the unit of time. */
            str3 = parse_netlim_key(str1, &key2);
            if (!str3 || str3 == str1) {
                bad = 1;
                break;
                }
            str1 = str3;

            /* Make use of the units. Convert to files or bytes per second. */
            switch (key1) {
            case netlim_file:
                if (key0 != netlim_tx1rate)
                    bad = 1;
                break;
            /* Convert to bytes. */
            case netlim_bit:
                if (key0 == netlim_tx2rate)
                    val /= 8;
                else
                    bad = 1;
                break;
            case netlim_byte:
                if (key0 != netlim_tx2rate)
                    bad = 1;
                break;
            default:
                bad = 1;
                break;
                } /* End of switch (key1). */

            if (bad)
                break;

            switch (key2) {
            /* Convert to seconds. */
            case netlim_sec:
                break;
            case netlim_min:
                val /= 60.0;
                break;
            case netlim_hr:
                val /= 3600.0;
                break;
            case netlim_day:
                val /= 86400.0;
                break;
            default:
                bad = 1;
                break;
                } /* End of switch (key2). */

            if (bad)
                break;

            /* Set the value. */
            if (key0 == netlim_tx1rate)
                obj1->tx1cred_rate = val;
            else if (key0 == netlim_tx2rate)
                obj1->tx2cred_rate = val;
            }
            break;
        case netlim_tx1max:
        case netlim_tx2max:
            {
            /* Expect something like "20.0 files" or "3000000 bytes". */
            netlim_enum key1 = netlim_MAX;
            const char* str3 = 0;
            char* str4 = 0;

            double val = strtod(str1, &str4);
            if (!str4 || str4 == str1 || errno == ERANGE) {
                bad = 1;
                break;
                }
            str1 = str4;

            /* Fetch units of the maximum, e.g. "files". */
            str3 = parse_netlim_key(str1, &key1);
            if (!str3 || str3 == str1) {
                bad = 1;
                break;
                }
            str1 = str3;

            /* Make use of the units. Convert to files or bytes. */
            switch (key1) {
            case netlim_file:
                if (key0 != netlim_tx1max)
                    bad = 1;
                break;
            /* Convert to bytes. */
            case netlim_bit:
                if (key0 == netlim_tx2max)
                    val /= 8;
                else
                    bad = 1;
                break;
            case netlim_byte:
                if (key0 != netlim_tx2max)
                    bad = 1;
                break;
            default:
                bad = 1;
                break;
                } /* End of switch (key1). */

            if (bad)
                break;

            /* Set the value. */
            if (key0 == netlim_tx1max)
                obj1->tx1debt_max = val;
            else if (key0 == netlim_tx2max)
                obj1->tx2debt_max = val;
            }
            break;
        default:
            bad = 1;
            break;
            } /* End of switch(key0). */

        if (bad) {
            retval = -6;
            break;
            }
        } /* End of while(*str1). */

    return retval;
    } /* End of function bws_net_parse. */

/*---------------------------------------------------------------------------
Add new network to networks array.
Return the integer index of the record which was used.
If not possible to store the data, return a negative integer.
---------------------------------------------------------------------------*/
/*----------------------*/
/*      bws_net_add     */
/*----------------------*/
/* Comment out this function to get rid of compiler warnings. */
#if 0
static int32 bws_net_add(bws_shm_t* x, uint32 address, uint32 mask,
        double tx1cred_rate, double tx1debt_max,
        double tx2cred_rate, double tx2debt_max) {
    bws_net_t* obj1 = 0;

    if (! x)
        return -1;
    if (x->n_networks >= N_NETWORKS)
        return -2;
    obj1 = &x->networks[x->n_networks];
    bws_net_init(obj1);

    obj1->address = address;
    obj1->mask = mask;

    obj1->tx1cred_rate = tx1cred_rate;
    obj1->tx1debt_max = tx1debt_max;
    obj1->tx2cred_rate = tx2cred_rate;
    obj1->tx2debt_max = tx2debt_max;
    x->n_networks += 1;

    return x->n_networks;
    } /* End of function bws_net_add. */
#endif /* 0 */

/*---------------------------------------------------------------------------
Add new network to networks array.
Return the integer index of the record which was used.
If not possible to store the data, return a negative integer.
---------------------------------------------------------------------------*/
/*----------------------*/
/*  bws_net_add_subnet  */
/*----------------------*/
static int32 bws_net_add_subnet(bws_shm_t* x, bws_net_t* subnet) {
    bws_net_t* obj1 = 0;

    if (!x || !subnet)
        return -1;
    if (x->n_networks >= N_NETWORKS)
        return -2;

    /* Check sanity of bandwidth limitations. */
    if (subnet->tx1cred_rate < 0)
        return -3;
    if (subnet->tx1debt_max < 0)
        return -4;
    if (subnet->tx2cred_rate < 0)
        return -5;
    if (subnet->tx2debt_max < 0)
        return -6;

    obj1 = &x->networks[x->n_networks];
    bws_net_init(obj1);

    obj1->address = subnet->address;
    obj1->mask = subnet->mask;

    obj1->tx1cred_rate = subnet->tx1cred_rate;
    obj1->tx1debt_max = subnet->tx1debt_max;
    obj1->tx2cred_rate = subnet->tx2cred_rate;
    obj1->tx2debt_max = subnet->tx2debt_max;
    x->n_networks += 1;

    return x->n_networks;
    } /* End of function bws_net_add_subnet. */

/*-----------------------------------------------------------------------------
sem_init() initializes the semaphores.
It is called for _every_ bwshare_init invocation, and the semaphores are
removed again by each server cleanup operation.
So they are re-created each time.
It would have been better if Apache provided a _final_ cleanup call, but maybe
this is not possible because the server does not know if the following init
operation is going to succeed.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*       sem_init       */
/*----------------------*/
static void sem_init(void) {
    union semun semctl_arg;     /* Argument for semctl call. */
    int ret = 0;                /* Generic function return value. */

    n_sem_init += 1;

    /* Need semaphores to protect the memory against concurrent access. */
#if IPC_KEYS_PRIVATE
    sem_id = semget(IPC_PRIVATE, 1, IPC_CREAT | 0666);
#else
    sem_id = semget(SEM_KEY_BASE, 1, IPC_CREAT | 0666);
#endif
    if (sem_id < 0) {
        perror("semget");
        exit(1);
        }

    /* Set all semaphore values to 1. */
    semctl_arg.val = 1;
    ret = semctl(sem_id, 0, SETVAL, semctl_arg);
    if (ret < 0) {
        perror("semctl");
        exit(1);
        }

    /* Initialize the semaphore operations. */
    semop_grab.sem_num = 0;         /* Only one resource to manage. */
    semop_grab.sem_op = -1;         /* Take away the resource. */
    semop_grab.sem_flg = SEM_UNDO;  /* Undo actions if process exits. */

    semop_free.sem_num = 0;         /* Only one resource to manage. */
    semop_free.sem_op = 1;          /* Give back the resource. */
    semop_free.sem_flg = SEM_UNDO;  /* Undo actions if process exits. */

#if WE_HAVE_UNIXD_H || BW_AP_VERSION_1
    /* Change the semaphore ownership to "nobody" or "wwwrun" etc. */
    if (geteuid() == 0) {
        struct semid_ds child_data;

        memset(&child_data, 0, sizeof(child_data));
#if BW_AP_VERSION_1
        child_data.sem_perm.uid = ap_user_id;
        child_data.sem_perm.gid = ap_group_id;
#endif
#if WE_HAVE_UNIXD_H
        child_data.sem_perm.uid = unixd_config.user_id;
        child_data.sem_perm.gid = unixd_config.group_id;
#endif
        child_data.sem_perm.mode = 0600;

        semctl_arg.buf = &child_data;
        ret = semctl(sem_id, 0, IPC_SET, semctl_arg);
        if (ret < 0) {
            perror("semctl");
            exit(1);
            }
        }
#endif
    } /* End of function sem_init. */

/*-----------------------------------------------------------------------------
sem_remove() is supposed to be called when Apache is about to exit.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*      sem_remove      */
/*----------------------*/
static void sem_remove(void) {
    union semun semctl_arg;     /* Argument for semctl call. */

    /* If no semaphores have been created, then don't remove them. */
    if (sem_id <= 0)
        return;

    /* Remove the semaphore set "sem_id". */
    semctl_arg.val = 0;         /* This value is ignored, probably. */
    /* Ignore the return value. */
    semctl(sem_id, 0, IPC_RMID, semctl_arg);
    sem_id = 0;
    } /* End of function sem_remove. */

/*-----------------------------------------------------------------------------
sem_grab() sets a semaphore so that the shared memory can be used.
If the nesting depth is exactly one, the semaphore is grabbed.
Otherwise the new depth of call nesting is recorded.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*      sem_grab        */
/*----------------------*/
static void sem_grab(void) {
    int ret = 0;

    if (sem_id == 0)
        return;

    /* Check the nesting depth. */
    sem_nesting_depth += 1;
    if (sem_nesting_depth != 1)
        return;

    /* Execute the resource-grab operation. */
    ret = semop(sem_id, &semop_grab, 1);
    if (ret < 0) {
        perror("semop");
        /* Where is this exit value defined? */
        exit(APEXIT_CHILDFATAL);
        }
    } /* End of function sem_grab. */

/*-----------------------------------------------------------------------------
sem_free() resets a semaphore so that the shared memory can be used by others.
If the depth of nesting makes the 1 -> 0 transition, the semaphore is freed.
Otherwise the new depth of call nesting is recorded.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
There is a small problem here that if sem_id is changed from zero to non-zero
or vice versa between a sem_grab abd sem_free call, then things will go wrong.
It's difficult to see exactly how to deal with that.
If the semaphore is allocated between these calls, after previously not being
allocated, then the code should work.
But obviously playing with sem_id within critical sections is a very bad idea.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*      sem_free        */
/*----------------------*/
static void sem_free(void) {
    int ret = 0;

    if (sem_id == 0)
        return;

    /* Check the nesting depth. */
    if (sem_nesting_depth <= 0) {
        /* Should never happen. Indicates coding error or new sem_id. */
        return;
        }
    /* 2008-1-31: This is very odd looking code!! Is it correct? */
    sem_nesting_depth -= 1;
    if (sem_nesting_depth != 0)
        return;

    /* Execute the resource-free operation. */
    ret = semop(sem_id, &semop_free, 1);
    if (ret < 0) {
        perror("semop");
        /* Where is this exit value defined? */
        exit(APEXIT_CHILDFATAL);
        }
    } /* End of function sem_free. */

/*-----------------------------------------------------------------------------
sem_value() reads the value of semaphore 0.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*      sem_value       */
/*----------------------*/
static int sem_value(void) {
    union semun semctl_arg;     /* Argument for semctl call. */
    int ret = 0;

    /* If no semaphores have been created, then don't remove them. */
    if (sem_id <= 0)
        return 0;

    /* Read the semaphore value. */
    semctl_arg.val = 0;         /* This value is ignored. */
    ret = semctl(sem_id, 0, GETVAL, semctl_arg);
    /* Luckily, no valid return value can be negative. */
    if (ret < 0) {
        perror("semctl");
        /* Where is this exit value defined. */
        exit(APEXIT_CHILDFATAL);
        }
    /* The non-negative value of the semaphore. */
    return ret;
    } /* End of function sem_value. */

/*-----------------------------------------------------------------------------
bws_init() is supposed to clear out the whole shared memory structure to
get it ready for use.
It should not refer to anything outside the bws_shm_t structure.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
The structure initialized here is supposed to record the historical behaviour
of all remote client hosts.
When the moving-average bit-rate of any host exceeds or equals a maximum limit,
it is supposed to be either pinched off (throttled), or temporarily refused
service, or completely cut off.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
A standard heap structure is implemented within the hosts database.
This is initially singly linked through the "next" field.
The concurrency operators sem_grab/sem_free are used just in case this function
is called when multiple processes are running.
Integer array indexes are used.
The special array index (uint16)(-1) == 65535 represents the null index.
The non-free records will be put in lists whose head-pointers are in hash-table
records.
It is required that N_REMOTE_HOSTS <= 65535.
Then N_REMOTE_HOSTS - 1 <= 65534 < NULL_HOST = (uint16)(-1) == 65535.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*       bws_init       */
/*----------------------*/
static void bws_init(bws_shm_t* x) {
    long i = 0;
    bws_host_t* h0 = 0;
    bws_hash_t* ha0 = 0;

    if (!x)
        return;
    /* +++ Begin Critical Section. +++ */
    sem_grab();
    x->start_time = time((time_t*)0);

    /* Link all records through "next" in descending order. */
    for (i = 0, h0 = x->hosts; i < N_REMOTE_HOSTS; ++i, ++h0) {
        bws_host_init(h0);

        /* Note: Assume that NULL_HOST == (uint16)(-1). Error otherwise. */
        h0->next = (uint16)(i - 1);
        }

    /* Initialize all hash-table head-pointers to the null index. */
    for (i = 0, ha0 = x->hashtable; i < N_HASH_ENTRIES; ++i, ++ha0) {
        ha0->head = NULL_HOST;
        }
    x->hosts_free = N_REMOTE_HOSTS - 1;
    x->hwm_hard = hwm_hard_deft;
    x->hwm_soft = hwm_soft_deft;
    x->lwm_hard = lwm_hard_deft;
    x->lwm_soft = lwm_soft_deft;

    x->n_trace = 0;
    x->n_info = 0;
    x->n_logs = 0;
    x->n_bytes = 0;

    x->tx1cred_rate = tx1cred_rate_deft;
    x->tx1debt_max = tx1debt_max_deft;
    x->tx2cred_rate = tx2cred_rate_deft;
    x->tx2debt_max = tx2debt_max_deft;

    x->n_files_warn = n_files_warn_deft;
    x->n_bytes_warn = n_bytes_warn_deft;
    x->idle_show = idle_show_deft;

    memset(x->networks, 0, sizeof(bws_net_t) * N_NETWORKS);
    x->n_networks = 0;

#if 0
    /* Ad hoc experiment. Address/mask = 58.6.65.64/0xfffffff0. */
    bws_net_add(x, 0x3A064140, 0xfffffff0, 0.2000, 50, 10000, 5000000);

    /* Googlebot. Address/mask = 66.249.66.65/0xffffff00. */
    bws_net_add(x, 0x42F94200, 0xffffff00, 0.4000, 100, 20000, 5000000);
#endif

    sem_free();
    /* --- End Critical Section. --- */
    } /* End of function bws_init. */

/*-----------------------------------------------------------------------------
host_sort_compare() compares two host_sort_t structures.
Want to sort an array in increasing order by the "value" field.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*   host_sort_compare  */
/*----------------------*/
static int host_sort_compare(const void* left, const void* right) {
    if (!left || !right)
        return 0;

    /* Sort first by "keep" field. */
    if (((host_sort_t*)left)->keep < ((host_sort_t*)right)->keep)
        return -1;
    if (((host_sort_t*)left)->keep > ((host_sort_t*)right)->keep)
        return 1;

    if (((host_sort_t*)left)->value < ((host_sort_t*)right)->value)
        return -1;
    if (((host_sort_t*)left)->value > ((host_sort_t*)right)->value)
        return 1;
    return 0;
    } /* End of function host_sort_compare. */

/*-----------------------------------------------------------------------------
bws_garbage_disposal() is supposed to reduce the number of records in the hosts
table by removing the least interesting items.
The return value is the number of host records which have been removed.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
This function should be covered by a sem_grab/sem_free pair.
A nestable call to these functions is made just in case.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
This function contains by far the longest critical section in the module.
All other child processes will block while this section is dealt with.
Therefore it must be kept as short as possible.
However, the garbage disposal is quite rare if N_REMOTE_HOSTS is large.
If N_REMOTE_HOSTS is small, then the duration should be quite short.
There is probably no real problem.
But if there are performance problems with very large loads, the problems
would probably be caused by this function.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Summary of algorithm.
If n_items <= hwm_soft, do nothing and return.
If n_items <= hwm_hard, try to remove items, but leave behind those items which
are still "interesting".
If n_items > hwm_hard, remove items down to at least hwm_hard, even if some
interesting items must be deleted.

When removing uninteresting items, remove them until n_items == lwm_soft.
I.e. should always leave at least lwm items if there were initially lwm items.
If there are too many interesting items, then leave up to lwm_hard items.

The high water marks indicate the desired table size _after_ garbage
collection.
-----------------------------------------------------------------------------*/
/*----------------------*/
/* bws_garbage_disposal */
/*----------------------*/
static int32 bws_garbage_disposal(bws_shm_t* base) {
    uint16 hwm_hard = 0;
    uint16 hwm_soft = 0;
    uint16 lwm_hard = 0;
    uint16 lwm_soft = 0;
    int32 n_removed = 0;
    int32 n_host_records = 0;
    host_sort_t* hosts_sort = 0;
    host_sort_t* s0 = 0;
    bws_host_t* h2 = 0;
    int i = 0;
    time_t t0 = time((time_t*)0);

    /* Should check that N_REMOTE_HOSTS > 0 if it becomes a variable. */
    if (!base)
        return 0;

    /* +++ Begin Critical Section. +++ */
    /* This is by far the longest critical section in the module. */
    sem_grab();

    hwm_hard = base->hwm_hard;
    hwm_soft = base->hwm_soft;
    lwm_hard = base->lwm_hard;
    lwm_soft = base->lwm_soft;

    /*-------------------------------------------------------------------------
    If the water marks are stupid, use more sensible values.
    These operations should be correct even if N_REMOTE_HOSTS == 1.
    It should not give stupid results if N_REMOTE_HOSTS == 0.
    -------------------------------------------------------------------------*/
    if (hwm_hard >= N_REMOTE_HOSTS && hwm_hard > 0)
        hwm_hard = N_REMOTE_HOSTS - 1;
    if (hwm_soft > hwm_hard)
        hwm_soft = hwm_hard;
    if (lwm_hard >= hwm_soft)
        lwm_hard = hwm_soft - 1;
    if (lwm_soft > lwm_hard)
        lwm_soft = lwm_hard;

    /* Calculate the number of occupied host records. */
    for (i = 0, h2 = base->hosts; i < N_REMOTE_HOSTS; ++i, ++h2) {
        /* Use the hash_index field to indicate "in use" status. */
        if (h2->hash_index != NULL_HASH)
            n_host_records += 1;
        }

    /* Do nothing if we're under the soft high-water mark. */
    if (n_host_records <= hwm_soft) {
        sem_free();
        return 0;
        }

    /*-------------------------------------------------------------------------
    Determine the level of interest of all host records.
    Use malloc because it isn't obvious which pool to stick this in.
    Should really check the return value here.
    -------------------------------------------------------------------------*/
    hosts_sort = (host_sort_t*)malloc(N_REMOTE_HOSTS * sizeof(host_sort_t));
    for (i = 0, s0 = hosts_sort, h2 = base->hosts;
            i < N_REMOTE_HOSTS;
            ++i, ++s0, ++h2
            ) {
        /* Link to a particular host record. */
        s0->host_index = i;

        /* Use the hash_index field to indicate "in use" status. */
        if (h2->hash_index == NULL_HASH) {
            s0->keep = 5;
            s0->value = 0;
            }
        else {
            /* Calculate the value of this host record. Very ad hoc!!!! */
            double value = 0;
            double dt = t0 - h2->last_time;

            /* Calculate the TX debt. But don't update the shm data. */
            double d1 = h2->tx1debt - h2->tx1cred_rate * dt;
            double d2 = h2->tx2debt - h2->tx2cred_rate * dt;

            if (d1 < 0)
                d1 = 0;
            if (d2 < 0)
                d2 = 0;

            /* If there is nothing really interesting, then keep = 0. */
            s0->keep = 0;

            /* If the last request was recent, then keep = 1. */
            if (dt <= base->idle_show) {
                value += 2000;
                s0->keep = 1;
                }
            if (dt <= 1000)
                value += (1000 - dt) * 10;

            /* If there's a yellow warning, then keep = 2. */
            if (h2->n_files > base->n_files_warn) {
                value += 50000;
                s0->keep = 2;
                }
            value += h2->n_files * 1000;

            if (h2->n_bytes > base->n_bytes_warn) {
                value += 50000;
                s0->keep = 2;
                }
            value += h2->n_bytes / 10;

            /* If there's a red warning, then keep = 3. */
            if (h2->tx1debtpeak > h2->tx1debt_max) {
                value += 100000;
                s0->keep = 3;
                }
            value += 1000 * h2->tx1debtpeak;

            if (h2->tx2debtpeak > h2->tx2debt_max) {
                value += 100000;
                s0->keep = 3;
                }
            value += h2->tx2debtpeak / 10;

            /* If there is any current debt, then keep = 4. */
            if (d1 > 0) {
                value += 1000000 + 100 * d1;
                s0->keep = 4;
                }

            if (d2 > 0) {
                value += 1000000 + d2;
                s0->keep = 4;
                }

            s0->value = value;
            }
        } /* End of for-loop to initialize hosts_sort array. */

    /* Sort the set of host records. */
    qsort(hosts_sort, N_REMOTE_HOSTS, sizeof(host_sort_t), host_sort_compare);

    /* Try to get the number of host records down to the low water mark. */
    /* Work through the sorted table to find records to remove. */
    for (i = 0, s0 = hosts_sort;
            i < N_REMOTE_HOSTS;
            ++i, ++s0
            ) {
        /* Break if sufficient records have been removed. */
        /* Records are sorted. So 'break' is equivalent to 'continue' here. */
        /* Keep deleting records until there are only lwm_hard items. */
        /* When <= lwm_hard records, only delete items with keep < 2. */
        if (n_host_records <= lwm_hard && s0->keep >= 2)
            break;

        /* Never delete down to less than lwm_soft records. */
        /* Let N = number of records with keep >= 2. */
        /* Then records remaining = max(lwm_soft, min(lwm_hard, N)). */
        if (n_host_records <= lwm_soft)
            break;

        /* Delete the host record. */
        {   /* Local scoping block. */
            /* Find the host record. */
            uint16 host_index = s0->host_index;
            uint16 hash_index = NULL_HASH;
            bws_host_t* h0 = 0;
            uint16 host0 = NULL_HOST;
            uint16 host1 = NULL_HOST;

            /* Sanity check. Should never happen. */
            if (host_index == NULL_HOST)
                break;

            /* Find the hash key for the head of the list. */
            h0 = &base->hosts[host_index];
            hash_index = h0->hash_index;

            /* Sanity check. Should never happen. */
            if (hash_index == NULL_HASH)
                break;

            /* Traverse the list of host records searching for the record. */
            for (host0 = base->hashtable[hash_index].head;
                    host0 != NULL_HOST;
                    host1 = host0, host0 = base->hosts[host0].next
                    ) {
                /* Sanity check. Should never break. */
                if (base->hosts[host0].hash_index != hash_index)
                    break;

                /* If this is the record we're looking for, unlink it. */
                if (host0 == host_index) {
                    /* Unlink the record. */
                    if (host1 == NULL_HOST) {
                        /* Case that this is the head of the list. */
                        base->hashtable[hash_index].head = h0->next;
                        }
                    else {
                        /* Case that this is _not_ the head of the list. */
                        base->hosts[host1].next = h0->next;
                        }

                    /*---------------------------------------------------------
                    Re-initialize host record and put in the free list.
                    This is in the middle of a critical section.
                    So really should not clear all fields here.
                    Should clear them when _allocating_ a new record instead.
                    It's all cleared in bws_newhost() already!
                    ---------------------------------------------------------*/
                    bws_host_init(h0);
                    h0->next = base->hosts_free;
                    base->hosts_free = host_index;
                    h0->hash_index = NULL_HASH;

                    /* Update the count of non-zero records. */
                    n_host_records -= 1;
                    n_removed += 1;

                    /* Break out of the list-traversal. */
                    break;
                    }
                } /* End of list-traversal loop. */
            /* At this point, the host record should have been unlinked. */
            /* If host was not found in the list, this implies corruption. */
            } /* End of local scoping block. */
        /* From here, go back and examine another hosts_sort element. */
        } /* End of hosts_sort table traversal. */

    /* If still not under the hard high water mark, force some removals. */
    /* .... */


    /* Give back the sort-table memory. */
    free(hosts_sort);

    /* --- End Critical Section. --- */
    sem_free();

    /* Return the number of items removed. */
    return n_removed;
    } /* End of function bws_garbage_disposal. */

/*-----------------------------------------------------------------------------
bws_newhost() should return a new host record, if possible.
The return value is NULL_HOST if no new record can be allocated.
Otherwise the index of the new record is returned.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
This function should be called when (1) the list for a given hash key is empty,
or (2) the list for the hash key does not contain the desired IP address.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Very important:
This function _must_ be called from within a sem_grab()/semfree() pair!
-----------------------------------------------------------------------------*/
/*----------------------*/
/*      bws_newhost     */
/*----------------------*/
static uint16 bws_newhost(bws_shm_t* base) {
    uint16 host_index = NULL_HOST;
    bws_host_t* h0 = 0;

    /* Check for irrational arguments. */
    if (!base)
        return NULL_HOST;

    /* Must not return from within the critical section. */
    /* +++ Begin Critical Section. +++ */
    sem_grab();

    /* Get the next free host record, if any. */
    host_index = base->hosts_free;
    if (host_index == NULL_HOST) {
        /*---------------------------------------------------------------------
        Case that the free list is exhausted.
        Here must retire one or more records, preferably the oldest.
        Should have a high water mark of number of hosts.
        Above this value, zero-debt host records are cleared until
        the record count reaches the low water mark, if possible.
        If no records can be cleared in this way, then the algorithm waits
        until there are no free records at all, and then a single
        record is cleared - the "least important" one.
        ---------------------------------------------------------------------*/
        bws_garbage_disposal(base);

        host_index = base->hosts_free;
        if (host_index == NULL_HOST) {
            /* This should never happen. But if it does, return cleanly. */

            sem_free();
            return NULL_HOST;
            }

        /* Initialize a recycled host record. */
        /* .... */

        }
    h0 = &base->hosts[host_index];

    /* Move the free-pointer to the next free host-record. */
    base->hosts_free = h0->next;

    /* Initialize the new host record. */
    /* Caller must change the "next", "hash_index", "address" fields. */
    /* Caller should change the "last_time" fields. */
    bws_host_init(h0);

    /* Set the bandwidth envelope parameters to the global default. */
    h0->tx1cred_rate = base->tx1cred_rate;
    h0->tx1debt_max = base->tx1debt_max;
    h0->tx2cred_rate = base->tx2cred_rate;
    h0->tx2debt_max = base->tx2debt_max;

    /* --- End Critical Section. --- */
    sem_free();

    return host_index;
    } /* End of function bws_newhost. */

/*-----------------------------------------------------------------------------
host_rdns_set() sets the reverse DNS translation in a given host record.
Important: Must use sem_grab() whenn calling this function!!!
An unusual byte storage format is used.
Instead of wasting a byte to store the terminal null, and to save any problems
with inefficient space usage due to word alignment, the first byte of the DNS
cache string is used for the number of bytes in the name string, and the rest
is a non-null-terminated string. Generally, there _may_ be a null termination,
but this should not be relied upon. Therefore use memcpy, not strcpy!
The argument rem_name of this function _is_ expected to be null-terminated.
The maximum string size that can be stored is N_DNS_CHARS - 1.
If this exceeded, leading bytes are discarded.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*     host_rdns_set    */
/*----------------------*/
static void host_rdns_set(bws_host_t* h0, const char* rem_name) {
    int n = 0;

    if (!h0)
        return;
    if (!rem_name || !*rem_name) {
        h0->rdns[0] = 0;
        return;
        }

    /* Copy the reverse DNS translation to the DNS cache field. */
    n = strlen(rem_name);
    if (n <= N_DNS_CHARS - 1) {
        /* The string fits. */
        memcpy(h0->rdns + 1, rem_name, n);
        }
    else {
        /* The string does not fit. */
        memcpy(h0->rdns + 1, rem_name + n - (N_DNS_CHARS - 1),
               N_DNS_CHARS - 1);
        }

    /* Set the number of bytes to the lesser of n and 255. */
    h0->rdns[0] = (n <= 255) ? n : 255;
    } /* End of function host_rdns_set. */

/*-----------------------------------------------------------------------------
compare_ip() compares two host_sort2_t structures.
Null host pointer structures are sorted high.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*      compare_ip      */
/*----------------------*/
static int compare_ip(const void* p1, const void* p2) {
    bws_host_t* h1;
    bws_host_t* h2;

    /* Check sanity of arguments. */
    if (!p1 || !p2)
        return 0;
    h1 = ((host_sort2_t*)p1)->host;
    h2 = ((host_sort2_t*)p2)->host;

    /* Case that a host pointer is null. */
    if (!h1)
        return h2 ? 1 : 0;
    if (!h2)
        return h1 ? -1 : 0;

    /* Compare the two host records. */
    if (h1->address < h2->address)
        return -1;
    if (h1->address > h2->address)
        return 1;
    return 0;
    } /* End of function compare_ip. */

/*-----------------------------------------------------------------------------
compare_idle() compares two host_sort2_t structures.
Null host pointer structures are sorted high.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*     compare_idle     */
/*----------------------*/
static int compare_idle(const void* p1, const void* p2) {
    bws_host_t* h1;
    bws_host_t* h2;

    /* Check sanity of arguments. */
    if (!p1 || !p2)
        return 0;
    h1 = ((host_sort2_t*)p1)->host;
    h2 = ((host_sort2_t*)p2)->host;

    /* Case that a host pointer is null. */
    if (!h1)
        return h2 ? 1 : 0;
    if (!h2)
        return h1 ? -1 : 0;

    /* Compare the two host records. */
    /* Notice that last_time is sorted opposite to the idle time. */
    if (h1->last_time < h2->last_time)
        return 1;
    if (h1->last_time > h2->last_time)
        return -1;
    return 0;
    } /* End of function compare_idle. */

/*-----------------------------------------------------------------------------
compare_tx() compares two host_sort2_t structures.
Null host pointer structures are sorted high.
This function uses a global variable for the
-----------------------------------------------------------------------------*/
/*----------------------*/
/*      compare_tx      */
/*----------------------*/
static int compare_tx(const void* p1, const void* p2) {
    bws_host_t* h1;
    bws_host_t* h2;

    /* Check sanity of arguments. */
    if (!p1 || !p2)
        return 0;
    h1 = ((host_sort2_t*)p1)->host;
    h2 = ((host_sort2_t*)p2)->host;

    /* Case that a host pointer is null. */
    if (!h1)
        return h2 ? 1 : 0;
    if (!h2)
        return h1 ? -1 : 0;

    /* Compare the two host records. */
    /* Sort into non-increasing order. */
    if (((host_sort2_t*)p1)->tx_debt < ((host_sort2_t*)p2)->tx_debt)
        return 1;
    if (((host_sort2_t*)p1)->tx_debt > ((host_sort2_t*)p2)->tx_debt)
        return -1;
    return 0;
    } /* End of function compare_tx. */

/*-----------------------------------------------------------------------------
compare_rdns() compares two host_sort2_t structures.
Null host pointer structures are sorted high.
This is the tricky one.
Sort increasing by top level domain, then sub-domain, sub-sub-domain etc.

This function is Work in Progress!!!!!
Not ready to use yet!
-----------------------------------------------------------------------------*/
/*----------------------*/
/*     compare_rdns     */
/*----------------------*/
static int compare_rdns(const void* p1, const void* p2) {
    bws_host_t* h1;
    bws_host_t* h2;
    const char* rdns1;
    const char* rdns2;
    const char* rdns1pos;
    const char* rdns2pos;
    const char* rdns1end;
    const char* rdns2end;
    int rdns1len;
    int rdns2len;

    /* Check sanity of arguments. */
    if (!p1 || !p2)
        return 0;
    h1 = ((host_sort2_t*)p1)->host;
    h2 = ((host_sort2_t*)p2)->host;

    /* Case that a host pointer is null. */
    if (!h1)
        return h2 ? 1 : 0;
    if (!h2)
        return h1 ? -1 : 0;

    /* Compare the two host records. */
    /* Sort non-decreasing by reverse DNS string. */
    rdns1 = h1->rdns + 1;
    rdns1len = ((int)(unsigned char)h1->rdns[0] & 0xff);
    if (rdns1len > N_DNS_CHARS - 1)
        rdns1len = N_DNS_CHARS - 1;
    rdns1end = rdns1 + rdns1len - 1;
    rdns1pos = rdns1end;

    rdns2 = h2->rdns + 1;
    rdns2len = ((int)(unsigned char)h2->rdns[0] & 0xff);
    if (rdns2len > N_DNS_CHARS - 1)
        rdns2len = N_DNS_CHARS - 1;
    rdns2end = rdns2 + rdns2len - 1;
    rdns2pos = rdns2end;

    /* Find the last dot. */
    while (rdns1len > 0 && (*rdns1pos != '.')) {
        rdns1pos -= 1;
        rdns1len -= 1;
        }
    while (rdns2len > 0 && (*rdns2pos != '.')) {
        rdns2pos -= 1;
        rdns2len -= 1;
        }

    /* Etc. etc. etc. */
    /* .... */

/*
    if (h1->rdns < h2->rdns)
        return 1;
    if (h1->rdns > h2->rdns)
        return -1;
*/
    return 0;
    } /* End of function compare_rdns. */

/*-----------------------------------------------------------------------------
compare_files() compares two host_sort2_t structures.
Null host pointer structures are sorted high.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*     compare_files    */
/*----------------------*/
static int compare_files(const void* p1, const void* p2) {
    bws_host_t* h1;
    bws_host_t* h2;

    /* Check sanity of arguments. */
    if (!p1 || !p2)
        return 0;
    h1 = ((host_sort2_t*)p1)->host;
    h2 = ((host_sort2_t*)p2)->host;

    /* Case that a host pointer is null. */
    if (!h1)
        return h2 ? 1 : 0;
    if (!h2)
        return h1 ? -1 : 0;

    /* Compare the two host records. */
    /* Sort non-increasing by number of files. */
    if (h1->n_files < h2->n_files)
        return 1;
    if (h1->n_files > h2->n_files)
        return -1;
    return 0;
    } /* End of function compare_files. */

/*-----------------------------------------------------------------------------
compare_bytes() compares two host_sort2_t structures.
Null host pointer structures are sorted high.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*     compare_bytes    */
/*----------------------*/
static int compare_bytes(const void* p1, const void* p2) {
    bws_host_t* h1;
    bws_host_t* h2;

    /* Check sanity of arguments. */
    if (!p1 || !p2)
        return 0;
    h1 = ((host_sort2_t*)p1)->host;
    h2 = ((host_sort2_t*)p2)->host;

    /* Case that a host pointer is null. */
    if (!h1)
        return h2 ? 1 : 0;
    if (!h2)
        return h1 ? -1 : 0;

    /* Compare the two host records. */
    /* Sort non-increasing by number of bytes. */
    if (h1->n_bytes < h2->n_bytes)
        return 1;
    if (h1->n_bytes > h2->n_bytes)
        return -1;
    return 0;
    } /* End of function compare_bytes. */

/*-----------------------------------------------------------------------------
compare_key() compares two host_sort2_t structures.
Null host pointer structures are sorted high.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*      compare_key     */
/*----------------------*/
static int compare_key(const void* p1, const void* p2) {
    bws_host_t* h1;
    bws_host_t* h2;

    /* Check sanity of arguments. */
    if (!p1 || !p2)
        return 0;
    h1 = ((host_sort2_t*)p1)->host;
    h2 = ((host_sort2_t*)p2)->host;

    /* Case that a host pointer is null. */
    if (!h1)
        return h2 ? 1 : 0;
    if (!h2)
        return h1 ? -1 : 0;

    /* Compare the two host records. */
    /* Sort non-decreasing by hash key. */
    if (h1->hash_index < h2->hash_index)
        return -1;
    if (h1->hash_index > h2->hash_index)
        return 1;
    return 0;
    } /* End of function compare_key. */

/*-----------------------------------------------------------------------------
hosts_sort2() sorts "hosts" for display.
The return value is a pool-allocated sorted host index array if successful, or
a null pointer otherwise.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*      hosts_sort2     */
/*----------------------*/
static host_sort2_t* hosts_sort2(
#if BW_AP_VERSION_1
        bws_shm_t* copy, host_sort_enum sort_type, pool* p) {
#else
        bws_shm_t* copy, host_sort_enum sort_type, apr_pool_t* p) {
#endif
    host_sort2_t* sorter = 0;
    host_sort2_t* s0 = 0;
    int i = 0;
    bws_host_t* h0 = 0;

    if (!copy || !p)
        return 0;

    /* Create an array sorter array. */
    /* (Perhaps this should be statically allocated.) */
#if BW_AP_VERSION_1
    sorter =
        (host_sort2_t*)ap_palloc(p, N_REMOTE_HOSTS * sizeof(host_sort2_t));
#else
    sorter =
        (host_sort2_t*)apr_palloc(p, N_REMOTE_HOSTS * sizeof(host_sort2_t));
#endif
    if (!sorter)
        return 0;

    /* Initialize the sorter array. */
    for (i = 0, s0 = sorter, h0 = copy->hosts;
            i < N_REMOTE_HOSTS;
            ++i, ++s0, ++h0
            ) {

        /* Use the hash_index field to indicate "in use" status. */
        if (h0->hash_index != NULL_HASH) {
            /* Link to a particular host record. */
            s0->host = h0;
            s0->host_index = i; /* This field probably not used. */
            }
        else {
            s0->host = 0;
            s0->host_index = NULL_HASH;
            }
        }

    /* Sort according to the indicated sorting style. */
    switch(sort_type) {
    case sort_ip:
        qsort(sorter, N_REMOTE_HOSTS, sizeof(host_sort2_t), compare_ip);
        break;
    case sort_idle:
        qsort(sorter, N_REMOTE_HOSTS, sizeof(host_sort2_t), compare_idle);
        break;
    case sort_tx1: {
        /* Get the current clock time. */
        time_t t0 = time((time_t*)0);

        /* Initialize the tx_debt fields. */
        for (i = 0, s0 = sorter; i < N_REMOTE_HOSTS; ++i, ++s0) {
            h0 = s0->host;
            if (!h0)
                continue;
            s0->tx_debt = h0->tx1debt
                        - (t0 - h0->last_time) * h0->tx1cred_rate;
            if (s0->tx_debt < 0)
                s0->tx_debt = 0;
            }
        qsort(sorter, N_REMOTE_HOSTS, sizeof(host_sort2_t), compare_tx);
        }
        break;
    case sort_tx2: {
        /* Get the current clock time. */
        time_t t0 = time((time_t*)0);

        /* Initialize the tx_debt fields. */
        for (i = 0, s0 = sorter; i < N_REMOTE_HOSTS; ++i, ++s0) {
            h0 = s0->host;
            if (!h0)
                continue;
            s0->tx_debt = h0->tx2debt
                        - (t0 - h0->last_time) * h0->tx2cred_rate;
            if (s0->tx_debt < 0)
                s0->tx_debt = 0;
            }
        qsort(sorter, N_REMOTE_HOSTS, sizeof(host_sort2_t), compare_tx);
        }
        break;
    case sort_tx1peak: {
        /* Initialize the tx_debt fields. */
        for (i = 0, s0 = sorter; i < N_REMOTE_HOSTS; ++i, ++s0) {
            h0 = s0->host;
            if (!h0)
                continue;
            s0->tx_debt = h0->tx1debtpeak;
            }
        qsort(sorter, N_REMOTE_HOSTS, sizeof(host_sort2_t), compare_tx);
        }
        break;
    case sort_tx2peak: {
        /* Initialize the tx_debt fields. */
        for (i = 0, s0 = sorter; i < N_REMOTE_HOSTS; ++i, ++s0) {
            h0 = s0->host;
            if (!h0)
                continue;
            s0->tx_debt = h0->tx2debtpeak;
            }
        qsort(sorter, N_REMOTE_HOSTS, sizeof(host_sort2_t), compare_tx);
        }
        break;
    case sort_rdns:
        /* Note: compare_rdns is not yet implemented. */
        qsort(sorter, N_REMOTE_HOSTS, sizeof(host_sort2_t), compare_rdns);
        break;
    case sort_files:
        qsort(sorter, N_REMOTE_HOSTS, sizeof(host_sort2_t), compare_files);
        break;
    case sort_bytes:
        qsort(sorter, N_REMOTE_HOSTS, sizeof(host_sort2_t), compare_bytes);
        break;
    case sort_key:
        qsort(sorter, N_REMOTE_HOSTS, sizeof(host_sort2_t), compare_key);
        break;
    default:
    case sort_index:
        /* Do nothing. */
        break;
        } /* End of switch(sort_type). */

    return sorter;
    } /* End of function hosts_sort2. */

/*-----------------------------------------------------------------------------
This function should be called once only, no matter how many times the server
is re-configured.
This function must set up some shared memory for recording the bandwidth usage
of all web clients.
Semaphores are set up for arbritrating access to this shared memory structure.
System V IPC is chosen because that's what I've seen in other people's modules.
If it was desired to combine the information from two or more running web
servers so as to control bandwidth over all servers, then non-private keys
should be used for memory and semaphores.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
The standard sequence for shared memory seems to be:
1.  create it
2.  attach it
3.  "remove" it     (i.e. mark it for removal-on-last-exit)
4.  initialize it
5.  use it
Then all child processes inherit everything as they should, and any other
running instances of this module will not clash.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
The corresponding sequence for semaphores is:
1.  create it
2.  initialize it
3.  use it
4.  remove it       (this is carried out immediately)
-----------------------------------------------------------------------------*/
/*----------------------*/
/*       shm_init       */
/*----------------------*/
static void shm_init(void) {
    int ret = 0;                /* Generic function return value. */

    n_shm_init += 1;

    /* Allocate and attach some memory. */
    /* Here should check to see if the key has been used already. */
#if IPC_KEYS_PRIVATE
    shm_id = shmget(IPC_PRIVATE, sizeof(bws_shm_t), IPC_CREAT | 0666);
#else
    shm_id = shmget(SHM_KEY_BASE, sizeof(bws_shm_t), IPC_CREAT | 0666);
#endif
    if (shm_id < 0) {
        perror("shmget");
        exit(1);
        }

    /* Attach the memory to this process. */
/*    shm_base = shmat(shm_id, (const void*)0, 0); */
    shm_base = shmat(shm_id, 0, 0);
    if (shm_base == (bws_shm_t*)(-1)) {
        perror("shmat");
        exit(1);
        }

#if WE_HAVE_UNIXD_H || BW_AP_VERSION_1
    /* Change the shared_memory ownership to "nobody" or "wwwrun" etc. */
    if (geteuid() == 0) {
        struct shmid_ds child_data;

        memset(&child_data, 0, sizeof(child_data));
#if BW_AP_VERSION_1
        child_data.shm_perm.uid = ap_user_id;
        child_data.shm_perm.gid = ap_group_id;
#endif
#if WE_HAVE_UNIXD_H
        child_data.shm_perm.uid = unixd_config.user_id;
        child_data.shm_perm.gid = unixd_config.group_id;
#endif
        child_data.shm_perm.mode = 0600;
        ret = shmctl(shm_id, IPC_SET, &child_data);
        if (ret < 0) {
            perror("shmctl");
            exit(1);
            }
        }
#endif

    /* Immediately "remove" the memory so that it will delete on exit. */
    /* This is okay because child processes still inherit the memory. */
    ret = shmctl(shm_id, IPC_RMID, (struct shmid_ds*)0);
    if (ret < 0) {
        perror("shmctl");
        exit(1);
        }

    /* Initialize the shared-memory data. */
    /* At this point, no other processes should be running. */
    bws_init(shm_base);
    } /* End of function shm_init. */

/*----------------------*/
/*      shm_remove      */
/*----------------------*/
static void shm_remove(void) {
    if (! shm_base)
        return;

    /* Ignore the return value. */
    shmdt(shm_base);
    shm_base = 0;
    } /* End of function shm_remove. */

/*-----------------------------------------------------------------------------
This function is called each time the parent process re-configures.
It is apparently called just before doing a reconfiguration.
So (n_init - n_plain_cleanup) should always equal 1 when displayed, and 0 just
before an "init" call.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
This would be a suitable place to reset all the shared memory and semaphores
and other global resources before the (probably) forthcoming "init" function
call.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
The version 2 return value is an integer.
-----------------------------------------------------------------------------*/
/*--------------------------*/
/*   plain_cleanup_handler  */
/*--------------------------*/
#if BW_AP_VERSION_1
static void plain_cleanup_handler(void* s) {
#else
static apr_status_t plain_cleanup_handler(void* s) {
#endif
    n_plain_cleanup += 1;
    sem_remove();
    shm_remove();
#if BW_AP_VERSION_2
    return APR_SUCCESS;
#endif
    } /* End of function plain_cleanup_handler. */

/*-----------------------------------------------------------------------------
Since this is only supposed to be called just before an "exec" in a child
process, it always seems that n_child_cleanup == 0.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
The version 2 return value is an integer.
-----------------------------------------------------------------------------*/
/*--------------------------*/
/*   child_cleanup_handler  */
/*--------------------------*/
#if BW_AP_VERSION_1
static void child_cleanup_handler(void* s) {
#else
static apr_status_t child_cleanup_handler(void* s) {
#endif
    n_child_cleanup += 1;
#if BW_AP_VERSION_2
    return APR_SUCCESS;
#endif
    } /* End of function child_cleanup_handler. */

/*-----------------------------------------------------------------------------
This function is called (probably once only) as the first call to the module.
Apparently it gets called twice for a DSO (Dynamic Shared Object?) module in
separate processes, of which one is immediately killed.
But for static modules, the call occurs twice in the same process.
So this is very inconvenient for creating shared memory etc.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Experiments show that this function is called twice when the program is
first run. It is run again when the program is re-configured with the
"kill -HUP" signal.
Presumably this means that shared memory can be initialised if and only
if n_init == 0 when this function is entered!??
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
In this function, the shared memory and semaphores must be created, and if/when
the function is called again, the memory/semaphores should be retained, but the
information should be updated, because the information might now be wrong.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*      bwshare_init    */
/*----------------------*/
#if BW_AP_VERSION_1
static void bwshare_init(server_rec* s, pool* p) {
#else
/* Where is the following set of args (for Apache 2.0.16) documented??!
   This function is replaced in Apache 2.0.44 by bwshare_module_init(),
   which returns an interger status code. See below.
*/
static void bwshare_init(apr_pool_t* p,
        apr_pool_t* plog, apr_pool_t* ptemp, server_rec* s) {
#endif
    n_init += 1;

    /* Initialize the shared memory the first time only. */
    /* 2008-1-31: Actually must re-initialize every time because shm_remove */
    /* called every time the process is closed and re-opened. */
    if (n_init - n_plain_cleanup <= 1)
        shm_init();

    /* Initialize the semaphore set only once. */
    /* 2008-1-31: It looks like the same thing applies to sem_init(). */
    /* Since there is no reliable Apache API documentation this all has to */
    /* be worked out by trial-and-error. So don't blame me if this is wrong! */
/*    if (n_sem_init <= 0) */
    if (n_init - n_plain_cleanup <= 1)
        sem_init();

#if BW_AP_VERSION_1
    /* Register a pair of cleanup handlers. */
    /* Why is this protected from alarms? See ap_alloc.h. */
    ap_block_alarms();
    ap_register_cleanup(p, (void*)s,
        plain_cleanup_handler, child_cleanup_handler);
    ap_unblock_alarms();
#else
    apr_pool_cleanup_register(p, (void*)s,
        plain_cleanup_handler, child_cleanup_handler);
#endif
    } /* End of function bwshare_init. */

/*-----------------------------------------------------------------------------
This is called for each virtual host, namely for the default host and for
each virtual host declared in the httpd.conf file.
-----------------------------------------------------------------------------*/
/*----------------------*/
/* bwshare_create_server*/
/*----------------------*/
#if BW_AP_VERSION_1
static void* bwshare_create_server(pool* p, server_rec* s) {
#else
static void* bwshare_create_server(apr_pool_t* p, server_rec* s) {
#endif
    bws_context_t* pc;

    n_server += 1;

    /* ap_pcalloc clears the new memory. ap_palloc does not. */
#if BW_AP_VERSION_1
    pc = (bws_context_t*)ap_palloc(p, sizeof(bws_context_t));
#else
    pc = (bws_context_t*)apr_palloc(p, sizeof(bws_context_t));
#endif
    bws_context_ctor(pc);
    pc->i = n_server;

#if BW_AP_VERSION_2
#if BW_INIT_NEW_LOCATION
    /* 2005-10-3: I don't know why this should be here. As soon as Apache 2 */
    /* gets some proper documentation for developers, I'll figure this out. */
    /* Leo said this works. So I'm doing it. Honi soit qui mal y pense! */
    bwshare_init(p, (apr_pool_t*)0, (apr_pool_t*)0, s);
#endif
#endif
    return (void*)pc;
    } /* End of function bwshare_create_server. */

/*---------------------------------------------------------------------------
Merge the configuration parameters of a child context into a parent context.
Apache 2.0.53 declaration in http_config.h:

void *(*merge_server_config) (apr_pool_t *p, void *base_conf, void *new_conf);

Apache 1.3.26 has this:

    void *(*merge_dir_config) (pool *p, void *base_conf, void *new_conf);
---------------------------------------------------------------------------*/
/*----------------------*/
/* bwshare_merge_server */
/*----------------------*/
#if BW_AP_VERSION_1
static void* bwshare_merge_server(pool* p,
        void* base_conf, void* new_conf) {
#else
static void* bwshare_merge_server(apr_pool_t* p,
        void* base_conf, void* new_conf) {
#endif
    bws_context_t *base_context, *new_context, *merged_context;

    base_context = (bws_context_t*)base_conf;
    new_context = (bws_context_t*)new_conf;
    /* Check for impossible craziness. */
    if (!base_context || !new_context)
        return 0;

#if BW_AP_VERSION_1
    merged_context = (bws_context_t*)ap_palloc(p, sizeof(bws_context_t));
#else
    merged_context = (bws_context_t*)apr_palloc(p, sizeof(bws_context_t));
#endif
    /* Check for possible craziness. */
    if (!merged_context)
        return 0;
    bws_context_ctor(merged_context);
    /* Do something silly with the useless field. */
    merged_context->i = (base_context->i << 4) | (new_context->i & 0x0f);

    /* For now, just inherit the child's throttle settings. */
    merged_context->throttle_off = new_context->throttle_off;

    return merged_context;
    } /* End of function bwshare_merge_server. */

/*---------------------------------------------------------------------------
Set a parameter from the httpd config file.
---------------------------------------------------------------------------*/
/*----------------------*/
/*     bw_tx1debt_max   */
/*----------------------*/
static const char* bw_tx1debt_max(cmd_parms* cmd, void* mconfig,
        const char* arg1) {
    double x = 0;
    char* tmp1 = 0;

    /* Reject uninitialised shared memory or bad argument. */
    if (! shm_base || ! arg1)
        return NULL;

    /* Parse the argument. */
    x = strtod(arg1, &tmp1);
    if (tmp1 != arg1 && x > 0)
        shm_base->tx1debt_max = x;
    return NULL;
    } /* End of function bw_tx1debt_max. */

/*---------------------------------------------------------------------------
Set a parameter from the httpd config file.
---------------------------------------------------------------------------*/
/*----------------------*/
/*    bw_tx1cred_rate   */
/*----------------------*/
static const char* bw_tx1cred_rate(cmd_parms* cmd, void* mconfig,
        const char* arg1) {
    double x = 0;
    char* tmp1 = 0;

    /* Reject uninitialised shared memory or bad argument. */
    if (! shm_base || ! arg1)
        return NULL;

    /* Parse the argument. */
    x = strtod(arg1, &tmp1);
    if (tmp1 != arg1 && x > 0)
        shm_base->tx1cred_rate = x;
    return NULL;
    } /* End of function bw_tx1cred_rate. */

/*---------------------------------------------------------------------------
Set a parameter from the httpd config file.
---------------------------------------------------------------------------*/
/*----------------------*/
/*     bw_tx2debt_max   */
/*----------------------*/
static const char* bw_tx2debt_max(cmd_parms* cmd, void* mconfig,
        const char* arg1) {
    double x = 0;
    char* tmp1 = 0;

    /* Reject uninitialised shared memory or bad argument. */
    if (! shm_base || ! arg1)
        return NULL;

    /* Parse the argument. */
    x = strtod(arg1, &tmp1);
    if (tmp1 != arg1 && x > 0)
        shm_base->tx2debt_max = x;
    return NULL;
    } /* End of function bw_tx2debt_max. */

/*---------------------------------------------------------------------------
Set a parameter from the httpd config file.
---------------------------------------------------------------------------*/
/*----------------------*/
/*    bw_tx2cred_rate   */
/*----------------------*/
static const char* bw_tx2cred_rate(cmd_parms* cmd, void* mconfig,
        const char* arg1) {
    double x = 0;
    char* tmp1 = 0;

    /* Reject uninitialised shared memory or bad argument. */
    if (! shm_base || ! arg1)
        return NULL;

    /* Parse the argument. */
    x = strtod(arg1, &tmp1);
    if (tmp1 != arg1 && x > 0)
        shm_base->tx2cred_rate = x;
    return NULL;
    } /* End of function bw_tx2cred_rate. */

/*---------------------------------------------------------------------------
Set a parameter from the httpd config file. Expecting an integer.
---------------------------------------------------------------------------*/
/*----------------------*/
/*    bw_throttle_off   */
/*----------------------*/
static const char* bw_throttle_off(cmd_parms* cmd, void* mconfig,
        const char* arg1) {
    long x = 0;
    char* tmp1 = 0;
    bws_context_t* context_srv = 0;

    /* Reject uninitialised shared memory or bad argument. */
    if (!cmd || !cmd->server || !cmd->server->module_config || !arg1)
        return NULL;
    context_srv = (bws_context_t*)
        ap_get_module_config(cmd->server->module_config, &bwshare_module);
    if (!context_srv)
        return NULL;

    /* Parse the argument. */
    x = strtol(arg1, &tmp1, 0);
    if (tmp1 != arg1)
        context_srv->throttle_off = x;
    return NULL;
    } /* End of function bw_throttle_off. */

/*---------------------------------------------------------------------------
Set a parameter from the httpd config file. Expecting a complex line.
This is a global setting, not a per-virtual-server setting.
---------------------------------------------------------------------------*/
/*----------------------*/
/*    bw_subnet_limit   */
/*----------------------*/
static const char* bw_subnet_limit(cmd_parms* cmd, void* mconfig,
        const char* args) {
#if 0
    bws_context_t* context_srv = 0;
#endif
    bws_net_t subnet1;
    int ret0 = 0;

    /* Set the bandwidth envelope parameters to the global default. */
    bws_net_init(&subnet1);
    /* Default subnet is everything! Alternatively make default nothing. */
    subnet1.address = 0x00000000;
    subnet1.mask = 0x00000000;
    if (shm_base) {
        subnet1.tx1cred_rate = shm_base->tx1cred_rate;
        subnet1.tx1debt_max = shm_base->tx1debt_max;
        subnet1.tx2cred_rate = shm_base->tx2cred_rate;
        subnet1.tx2debt_max = shm_base->tx2debt_max;
        }
    else {
        subnet1.tx1cred_rate = tx1cred_rate_deft;
        subnet1.tx1debt_max = tx1debt_max_deft;
        subnet1.tx2cred_rate = tx2cred_rate_deft;
        subnet1.tx2debt_max = tx2debt_max_deft;
        }

    /* Reject uninitialised shared memory or bad argument. */
    if (!cmd || !cmd->server || !cmd->server->module_config || !args)
        return NULL;
#if 0
    /* Fetch the virtual-server context. */
    context_srv = (bws_context_t*)
        ap_get_module_config(cmd->server->module_config, &bwshare_module);
    if (!context_srv)
        return NULL;
#endif /* 0 */

    /* Experiment to make sure we're getting the right text. */
#if 0
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, cmd->server,
        "args = \"%s\"", args);
#endif

    /* Parse the argument. */
    ret0 = bws_net_parse(&subnet1, args);

#if 0
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, cmd->server,
        "address = 0x%08lx, mask = 0x%08lx"
        ", tx1rate = %.03lf, tx1max = %.0lf, tx2rate = %.0lf, tx2max = %.0lf",
        (unsigned long)subnet1.address, (unsigned long)subnet1.mask,
        subnet1.tx1cred_rate, subnet1.tx1debt_max,
        subnet1.tx2cred_rate, subnet1.tx2debt_max);
#endif /* 0 */

    if (ret0 < 0)
        return NULL;

    bws_net_add_subnet(shm_base, &subnet1);
    return NULL;
    } /* End of function bw_subnet_limit. */

/*-----------------------------------------------------------------------------
Trace internal parameters and variables in HTML format.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*    trace_internal    */
/*----------------------*/
static int trace_internal(request_rec* r) {
    ap_rputs("\n<p><hr><b>child-process statistics</b>\n", r);
    ap_rputs("<p><table rules=all border=1 cellspacing=0>\n", r);
    ap_rprintf(r,
        "<tr><td>n_shm_init</td><td>%lu</td></tr>\n", n_shm_init);
    ap_rprintf(r,
        "<tr><td>n_sem_init</td><td>%lu</td></tr>\n", n_sem_init);
    ap_rprintf(r,
        "<tr><td>n_init</td><td>%lu</td></tr>\n", n_init);
    ap_rprintf(r,
        "<tr><td>n_server</td><td>%lu</td></tr>\n", n_server);
    ap_rprintf(r,
        "<tr><td>n_plain_cleanup</td><td>%lu</td></tr>\n", n_plain_cleanup);
    ap_rprintf(r,
        "<tr><td>n_child_cleanup</td><td>%lu</td></tr>\n", n_child_cleanup);
    ap_rprintf(r,
        "<tr><td>n_info</td><td>%lu</td></tr>\n", n_info);
    ap_rprintf(r,
        "<tr><td>n_trace</td><td>%lu</td></tr>\n", n_trace);
    ap_rprintf(r,
        "<tr><td>sem_value</td><td>%d</td></tr>\n", (int)sem_value());

    ap_rprintf(r,
        "<tr><td>sizeof(bws_shm_t)</td><td>%d</td></tr>\n",
        (int)sizeof(bws_shm_t));
    ap_rprintf(r,
        "<tr><td>sizeof(bws_host_t)</td><td>%d</td></tr>\n",
        (int)sizeof(bws_host_t));
    ap_rprintf(r,
        "<tr><td>sizeof(bws_hash_t)</td><td>%d</td></tr>\n",
        (int)sizeof(bws_hash_t));
    ap_rprintf(r,
        "<tr><td>sizeof(bws_net_t)</td><td>%d</td></tr>\n",
        (int)sizeof(bws_net_t));
    ap_rprintf(r,
        "<tr><td>n_remote_hosts</td><td>%d</td></tr>\n",
        (int)N_REMOTE_HOSTS);
    ap_rprintf(r,
        "<tr><td>n_hash_entries</td><td>%d</td></tr>\n",
        (int)N_HASH_ENTRIES);
    ap_rprintf(r,
        "<tr><td>n_networks</td><td>%d</td></tr>\n",
        (int)N_NETWORKS);
#if BW_AP_VERSION_1
    ap_rprintf(r, "<tr><td>user id</td><td>%d</td></tr>\n",
        (int)ap_user_id);
    ap_rprintf(r, "<tr><td>group id</td><td>%d</td></tr>\n",
        (int)ap_group_id);
#endif
#if WE_HAVE_UNIXD_H
    ap_rprintf(r, "<tr><td>user id</td><td>%d</td></tr>\n",
        (int)unixd_config.user_id);
    ap_rprintf(r, "<tr><td>group id</td><td>%d</td></tr>\n",
        (int)unixd_config.group_id);
#endif
#if BW_CONN_TRACE
    ap_rprintf(r, "<tr><td>handler family<td>%d\n",
        (int)ctt_handler.family);
    ap_rprintf(r, "<tr><td>handler port<td>%d\n",
        (int)ctt_handler.port);
    ap_rprintf(r, "<tr><td>handler salen<td>%d\n",
        (int)ctt_handler.salen);
    ap_rprintf(r, "<tr><td>handler ipaddr_len<td>%d\n",
        (int)ctt_handler.ipaddr_len);
    ap_rprintf(r, "<tr><td>handler addr_str_len<td>%d\n",
        (int)ctt_handler.addr_str_len);

    ap_rprintf(r, "<tr><td>logger family<td>%d\n",
        (int)ctt_logger.family);
    ap_rprintf(r, "<tr><td>logger port<td>%d\n",
        (int)ctt_logger.port);
    ap_rprintf(r, "<tr><td>logger salen<td>%d\n",
        (int)ctt_logger.salen);
    ap_rprintf(r, "<tr><td>logger ipaddr_len<td>%d\n",
        (int)ctt_logger.ipaddr_len);
    ap_rprintf(r, "<tr><td>logger addr_str_len<td>%d\n",
        (int)ctt_logger.addr_str_len);
#endif
    ap_rputs("</table>\n", r);
    } /* End of function trace_internal. */

/*-----------------------------------------------------------------------------
From the point of view of the bwshare module, the ability to serve up
information through file-type handlers like "bwshare_trace" is useful only
for management of the module.
Clearly it is straightforward to display stuff to a browser.
It may be a little more difficult to send instructions (in a php-like manner)
from a browser to the module to alter parameters.
Initially, control of the module will be through the httpd.conf file and
other configuration files referred to in httpd.conf.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Software for text I/O (parsing and code generation) is generally difficult to
write in a crisp, clear way. So this function is not a great work of art.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Command line parameters param=X:
rdns    X   if (X != 0), display the reverse DNS translations
set     X   if (X >= 1), show only "interesting" host records
trace   X   if (X >= 1), show child process info
recent  X   if (X >= 0), host records with idle time <= X are "interesting"
-----------------------------------------------------------------------------*/
/*----------------------*/
/*     bwshare_trace    */
/*----------------------*/
static int bwshare_trace(request_rec* r) {
    time_t t1 = 0;
    int display_rdns = 1;   /* Default is to do the DNS now. */
    int display_limits = 0; /* Default is to not show bandwidth limits. */
    int display_set = 1;    /* Default is to show only indebted hosts. */
    int trace_level = 0;    /* Trace level for this handler. */
    int idle_show = -1;     /* Recentness of accesses to display. */
    host_sort_enum sort_type = sort_idle;   /* Sort field. */

#if BW_AP_VERSION_2
    if (strcmp(r->handler, "bwshare-trace") != 0)
        return DECLINED;
#endif

    /* Try to work out how often this child process is called. */
    n_trace += 1;

    /* Dodgy non-protected access to shared memory. */
    if (shm_base) {
        /* Protect access to shared memory with a semaphore. */
        /* +++ Begin Critical Section. +++ */
        sem_grab();
        shm_base->n_trace += 1;
        sem_free();
        /* --- End Critical Section. --- */
        }

    /* Stick a header on the output first. */
    r->content_type = "text/html";

    /* Why is the header forced out here? */
#if BW_AP_VERSION_1
    ap_send_http_header(r);
#endif

    /* Is this for the case that "HEAD" is the action requested? */
    if (r->header_only)
        return OK;

    /* Now try to read the command-line arguments for this request. */
    if (r->args && *r->args) {
        char* args = r->args;
        while (args && *args) {
            char* arg1 = 0;
            char* arg2 = 0;

            /* Get the first word of the remaining string. */
            arg1 = ap_getword_nc(r->pool, &args, '&');
            /* ap_getword skips multiple matches. So null-string means end. */
            if (!arg1 || !*arg1)
                break;

            /* Get the string for the variable to set. */
            /* After this call, have arg2 '=' arg1. */
            arg2 = ap_getword_nulls_nc(r->pool, &arg1, '=');

            if ((strcmp(arg2, "rdns")) == 0) {
                /* Parameter of form "rdns=X". */
                display_rdns = atoi(arg1);
                }
            else if ((strcmp(arg2, "limits")) == 0) {
                /* Parameter of form "limits=X". */
                display_limits = atoi(arg1);
                }
            else if ((strcmp(arg2, "set")) == 0) {
                /* Parameter of form "set=X". */
                display_set = atoi(arg1);
                }
            else if ((strcmp(arg2, "trace")) == 0) {
                /* Parameter of form "trace=X". */
                trace_level = atoi(arg1);
                }
            else if ((strcmp(arg2, "recent")) == 0) {
                /* Parameter of form "recent=X". */
                idle_show = atoi(arg1);
                }
            else if ((strcmp(arg2, "sort")) == 0) {
                /* Parameter of form "sort=X". */
                if (strcmp(arg1, "ip") == 0)
                    sort_type = sort_ip;
                else if (strcmp(arg1, "idle") == 0)
                    sort_type = sort_idle;
                else if (strcmp(arg1, "files") == 0)
                    sort_type = sort_files;
                else if (strcmp(arg1, "bytes") == 0)
                    sort_type = sort_bytes;
                else if (strcmp(arg1, "tx1") == 0)
                    sort_type = sort_tx1;
                else if (strcmp(arg1, "tx2") == 0)
                    sort_type = sort_tx2;
                else if (strcmp(arg1, "tx1peak") == 0)
                    sort_type = sort_tx1peak;
                else if (strcmp(arg1, "tx2peak") == 0)
                    sort_type = sort_tx2peak;
                else if (strcmp(arg1, "rdns") == 0)
                    sort_type = sort_rdns;
                else if (strcmp(arg1, "key") == 0)
                    sort_type = sort_key;
                else if (strcmp(arg1, "index") == 0)
                    sort_type = sort_index;
                }
            /* Silently ignore any unrecognized parameters. */
            }
        }

    /* Some content. */
    t1 = time((time_t*)0);
    ap_rputs(
        "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
        "<html lang=\"en\"><head>\n"
     "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">\n"
        "<meta name=\"robots\" content=\"noindex,nofollow\">\n"
        "<title>bwshare " BWSHARE_VERSION " trace</title></head>\n"
        "<body bgcolor=\"" COLOUR_TRACE_BG "\"><center>\n"
        "<font size=\"+2\"><b>"
        "bwshare " BWSHARE_VERSION " trace</b><br></font>\n"
        "<a href=\"bwshare-info\">bwshare top menu page</a><br>\n", r);
    if (display_rdns <= 0) {
        if (display_set <= 0) {
            ap_rputs("<a href=\"bwshare-trace?set=0\">"
                "show reverse DNS translation</a><br>\n"
                "<a href=\"bwshare-trace?rdns=0\">"
                "show interesting host records only</a><br>\n", r);
            }
        else {
            ap_rputs("<a href=\"bwshare-trace\">"
                "show reverse DNS translation</a><br>\n"
                "<a href=\"bwshare-trace?rdns=0&set=0\">"
                "show all host records</a><br>\n", r);
            }
        }
    else {
        if (display_set <= 0) {
            ap_rputs("<a href=\"bwshare-trace?rdns=0&set=0\">"
                "hide reverse DNS translation</a><br>\n"
                "<a href=\"bwshare-trace\">"
                "show interesting host records only</a><br>\n", r);
            }
        else {
            ap_rputs("<a href=\"bwshare-trace?rdns=0\">"
                "hide reverse DNS translation</a><br>\n"
                "<a href=\"bwshare-trace?set=0\">"
                "show all host records</a><br>\n", r);
            }
        }
    ap_rputs("<a href=\"http://www.topology.org/src/bwshare/README.html\">"
        "source for this Apache module</a>\n", r);

    if (shm_base) {
        int i = 0;
        int j = 0;
        bws_shm_t* shm_copy = 0;
        time_t t0 = 0;
#if BW_AP_VERSION_2
        apr_time_t t0_64;
        apr_time_t t1_64;
#endif
        int n_host_records = 0;
        int n_hash_keys = 0;
        bws_host_t* h0 = 0;
        host_sort2_t* sorter = 0; /* Pool-allocated array for sorting hosts. */
        host_sort2_t* hs0 = 0;

#if BW_AP_VERSION_1
        ap_block_alarms();  /* I don't know if this is needed. */
#endif
        /*---------------------------------------------------------------------
        It seems a little risky to do I/O while lowering the semaphore for
        shared memory. So copy all of it in haste, and print at leisure.
        ---------------------------------------------------------------------*/
#if BW_AP_VERSION_1
        shm_copy = (bws_shm_t*)ap_palloc(r->pool, sizeof(bws_shm_t));
#else
        shm_copy = (bws_shm_t*)apr_palloc(r->pool, sizeof(bws_shm_t));
#endif
        if (! shm_copy) {
            /* This is a bad but unlikely error. */
            ap_rputs(
                "<p><hr>Error: allocation of memory for shm_copy failed!\n"
                "</body></html>\n", r);
#if BW_AP_VERSION_1
            ap_unblock_alarms(); /* Important!! */
#endif
            return OK;
            }

        /* Get a quick copy of the entire shared memory structure. */
        /* +++ Begin Critical Section. +++ */
        sem_grab();
        memcpy(shm_copy, shm_base, sizeof(bws_shm_t));
        sem_free();
        /* --- End Critical Section. --- */

        t0 = shm_copy->start_time;
#if BW_AP_VERSION_2
        apr_time_ansi_put(&t0_64, t0);
        apr_time_ansi_put(&t1_64, t1);
#endif
        ap_rputs("\n<p>"
            "<table rules=all border=1 cellspacing=0 cellpadding=2 bgcolor=\""
            COLOUR_TRACE_TAB0 "\">\n", r);

        /* Row 1. */
        ap_rprintf(r, "<tr><td>httpd uptime</td>"
            "<td align=right>%ld sec</td><td>&nbsp;<br></td>\n",
            (long)(t1 - t0));
#if BW_AP_VERSION_1
        ap_rprintf(r, "<td>start utc</td><td align=right>%s</td></tr>\n",
            ap_ht_time(r->pool, t0, "%Y-%m-%d %H:%M:%S", 1));
#else
        ap_rprintf(r, "<td>start utc</td><td align=right>%s</td></tr>\n",
            ap_ht_time(r->pool, t0_64, "%Y-%m-%d %H:%M:%S", 1));
#endif

        /* Row 2. */
        ap_rprintf(r, "<tr><td>average bitrate</td>"
            "<td align=right>%.0f bits/sec</td><td>&nbsp;<br></td>\n",
            (t1 > t0) ? (((double)shm_copy->n_bytes * 8)/(t1 - t0)) : 0);
#if BW_AP_VERSION_1
        ap_rprintf(r,
            "<td>current utc</td><td align=right>%s</td></tr>\n",
            ap_ht_time(r->pool, t1, "%Y-%m-%d %H:%M:%S", 1));
#else
        ap_rprintf(r,
            "<td>current utc</td><td align=right>%s</td></tr>\n",
            ap_ht_time(r->pool, t1_64, "%Y-%m-%d %H:%M:%S", 1));
#endif

        /* Row 3. */
        ap_rprintf(r, "<tr><td>total requests</td>"
            "<td align=right>%ld files</td>\n",
            shm_copy->n_logs);
        /* This is a 64-bit integer! Type %qd was introduced in 1.3.7. */
        /* Why does %lld not work? */
        ap_rprintf(r, "<td>&nbsp;<br></td>"
            "<td>total data sent</td><td align=right>%s bytes</td></tr>\n",
            uint64comma_string(shm_copy->n_bytes));

        /* Row 4. */
        ap_rprintf(r, "<tr><td>tx1 credit rate</td><td align=right>"
            "%.02f files/min</td>\n", shm_copy->tx1cred_rate * 60);
        ap_rprintf(r, "<td>&nbsp;<br></td><td>tx2 credit rate</td>"
            "<td align=right>%lu bytes/sec</td></tr>\n",
            (unsigned long)shm_copy->tx2cred_rate);

        /* Row 5. */
        ap_rprintf(r,
            "<tr><td>tx1 debt max</td><td align=right>%.02f files</td>\n",
            shm_copy->tx1debt_max);
        ap_rprintf(r, "<td>&nbsp;<br></td><td>tx2 debt max</td>"
            "<td align=right>%lu bytes</td></tr>\n",
            (unsigned long)shm_copy->tx2debt_max);

        /* Row 6. */
        ap_rprintf(r,
            "<tr><td>files warning</td><td align=right>%ld files</td>\n",
            shm_copy->n_files_warn);
        ap_rprintf(r, "<td>&nbsp;<br></td><td>bytes warning</td>"
            "<td align=right>%ld bytes</td></tr>\n",
            shm_copy->n_bytes_warn);

        ap_rputs("</table>\n", r);

        ap_rputs("\n<p><b>client traffic statistics</b><br>\n", r);
        /* Flush output, in case the table takes a long time to create. */
        ap_rflush(r);

        /*---------------------------------------------------------------------
        Create an array of pointers to host records.
        Everything after this must not depend on the linkage to and
        from the hash table copy.
        ---------------------------------------------------------------------*/
        sorter = hosts_sort2(shm_copy, sort_type, r->pool);
        if (!sorter) {
            /* This is a bad but unlikely error. */
            ap_rputs(
                "<p><hr>Error: allocation of memory for hosts sorter failed!\n"
                "</body></html>\n", r);
#if BW_AP_VERSION_1
            ap_unblock_alarms(); /* Important!! */
#endif
            return OK;
            }

        /* Display a table of host records. */
        ap_rputs(
            "<table rules=all border=1 cellspacing=0 cellpadding=2 bgcolor=\""
            COLOUR_TRACE_TAB1 "\">\n<tr>", r);
        ap_rputs("<th>ip address</th>\n<th>files</th><th>bytes</th>"
            "<th>idle<br>time</th>\n"
            "<th>tx1debt</th><th>peak<br>tx1debt</th>\n"
            "<th>tx2debt</th><th>peak<br>tx2debt</th>\n", r);
        if (display_rdns)
            ap_rputs("<th>reverse DNS</th>\n", r);
        if (display_limits)
            ap_rputs("<th>tx1<br>rate</th><th>tx1<br>max</th>"
                "<th>tx2<br>rate</th><th>tx2<br>max</th>\n", r);
        if (show_key)
            ap_rputs("<th>key</th>", r);
        if (show_index)
            ap_rputs("<th>index</th>", r);
        ap_rputs("</tr>\n\n", r);

        /* Traverse the "hosts" table and display each record. */
        /* The index "i" is the hosts table index only if it is not sorted. */
/*        for (i = 0, h0 = shm_copy->hosts; i < N_REMOTE_HOSTS; ++i, ++h0) { */

        for (j = 0, hs0 = sorter; j < N_REMOTE_HOSTS; ++j, ++hs0) {
            /* Many local variable declarations. This is why I prefer C++. */
            unsigned long a = 0;
            unsigned long f = 0;
            unsigned long n = 0;
            double dt = 0;
            uint16 h = 0;
            double d1 = 0;
            double d2 = 0;
            double d1max = 0;
            double d2max = 0;
            char* dns_rev = 0;

            h0 = hs0->host;
            i = hs0->host_index;
            if (!h0)
                continue;
            /*-----------------------------------------------------------------
            Skip empty host records. A single field h0->hash_index indicates
            whether a host record is occupied or not.
            -----------------------------------------------------------------*/
            if (h0->hash_index == NULL_HASH)
                continue;

            /* Get some fields out of the host record. */
            a = (unsigned long)h0->address;
            f = (unsigned long)h0->n_files;
            n = (unsigned long)h0->n_bytes;
            h = h0->hash_index;
            dt = ((t1 >= h0->last_time) ? (t1 - h0->last_time) : 0);
            d1max = h0->tx1debtpeak;
            d2max = h0->tx2debtpeak;

            /* Recalculate the TX debt. But don't update the shm data. */
            /* Alternatively, could update all txdebt values while copying. */
            d1 = h0->tx1debt - h0->tx1cred_rate * dt;
            if (d1 < 0)
                d1 = 0;
            d2 = h0->tx2debt - h0->tx2cred_rate * dt;
            if (d2 < 0)
                d2 = 0;

            /* Take idle_show from command line if non-negative. */
            if (idle_show < 0)
                idle_show = shm_copy->idle_show;

            /* If set >= 1, show only hosts with debt or high peak debt. */
            /* These conditions should be used also in garbage collection. */
            if (display_set >= 1 && d1 <= 0 && d2 <= 0
                   && d1max <= h0->tx1debt_max
                   && d2max <= h0->tx2debt_max
                   && n <= shm_copy->n_bytes_warn
                   && f <= shm_copy->n_files_warn
                   && dt > idle_show)
                continue;

            /*--------------------------------------------------------------
            Look up the reverse DNS translation with verification.
            This can't be done here. It must be done when the request occurs,
            if the ap_get_remote_host() function is to be used.
            Otherwise must use gethostbyaddr() at this point.
            --------------------------------------------------------------*/
            if (display_rdns) {
                /* This is the slow way to look up the reverse DNS. */
/*
                unsigned long a_rev = htonl(a);
                struct hostent* ph0
                    = gethostbyaddr((const char*)&a_rev, 4, AF_INET);
                if (ph0 && ph0->h_name)
                    dns_rev = ph0->h_name;
                else
                    dns_rev = "<br>";
*/
                /* Copy the cached DNS address. This should be in a function!*/
                /* Format of string: length-byte + non-null-terminated str. */
                int n_rdns_bytes = ((int)(unsigned char)h0->rdns[0] & 0xff);

                if (n_rdns_bytes > 0) {
                    /* Number of bytes to copy from cache string. */
                    /* Clever little algorithm here. */
                    int n_rdns_copy = n_rdns_bytes;
                    int n_rdns_hatch = n_rdns_copy - (N_DNS_CHARS - 1);
                    if (n_rdns_hatch > 0)
                        n_rdns_copy -= n_rdns_hatch;
                    else
                        n_rdns_hatch = 0;

                    /* Should check error return, but don't bother for now. */
#if BW_AP_VERSION_1
                    dns_rev = (char*)ap_palloc(r->pool,
                              1 + n_rdns_bytes);
#else
                    dns_rev = (char*)apr_palloc(r->pool,
                              1 + n_rdns_bytes);
#endif
                    /* Prepend hatches, copy string, and null terminate. */
                    if (n_rdns_hatch > 0)
                        memset(dns_rev, '?', n_rdns_hatch);
                    memcpy(dns_rev + n_rdns_hatch, &h0->rdns[1], n_rdns_copy);
                    /* Null termination. */
                    dns_rev[n_rdns_bytes] = 0;
                    }
                else {
                    /* Default string for absent DNS translation. */
                    dns_rev = "<br>";
                    }
                }

            /* Display the IP address, n files, n bytes and idle time. */
            ap_rprintf(r, "<tr><td align=right>%s</td>\n", ipstring(a));

            /* Display current number of requests. */
            if (f > shm_copy->n_files_warn)
                ap_rputs("<td align=right bgcolor=\"" COLOUR_TRACE_WARN "\">",
                    r);
            else
                ap_rputs("<td align=right>", r);
            ap_rprintf(r, "%lu</td>", f);

            /* Display current number of bytes transmitted. */
            if (n > shm_copy->n_bytes_warn)
                ap_rputs("<td align=right bgcolor=\"" COLOUR_TRACE_WARN "\">",
                    r);
            else
                ap_rputs("<td align=right>", r);
            ap_rprintf(r, "%lu</td>", n);

            /* Display the idle time - the time since last request log. */
            ap_rprintf(r, "<td align=right>%lu</td>\n", (unsigned long)dt);

            /* Display current TX 1 debt. */
            if (d1 > h0->tx1debt_max)
                ap_rputs("<td align=right bgcolor=\"" COLOUR_TRACE_BAD "\">",
                    r);
            else
                ap_rputs("<td align=right>", r);
            ap_rprintf(r, "%.02f</td>", d1);

            /* Display current TX 1 debt max. */
            if (d1max > h0->tx1debt_max)
                ap_rputs("<td align=right bgcolor=\"" COLOUR_TRACE_BAD "\">",
                    r);
            else
                ap_rputs("<td align=right>", r);
            ap_rprintf(r, "%.02f</td>\n", d1max);

            /* Display current TX 2 debt. */
            if (d2 > h0->tx2debt_max)
                ap_rputs("<td align=right bgcolor=\"" COLOUR_TRACE_BAD "\">",
                    r);
            else
                ap_rputs("<td align=right>", r);
            ap_rprintf(r, "%lu</td>", (unsigned long)d2);

            /* Display current TX 2 debt max. */
            if (d2max > h0->tx2debt_max)
                ap_rputs("<td align=right bgcolor=\"" COLOUR_TRACE_BAD "\">",
                    r);
            else
                ap_rputs("<td align=right>", r);
            ap_rprintf(r, "%lu</td>\n", (unsigned long)d2max);

            /* Display the reverse domain name translation. */
            if (display_rdns)
                ap_rprintf(r, "<td>%s</td>\n", dns_rev);

            /* Display the bandwidth limits for the IP address. */
            if (display_limits) {
                /* Files/minute. */
                ap_rprintf(r, "<td align=right>%.02lf</td>\n",
                    h0->tx1cred_rate * 60);
                /* Files. */
                ap_rprintf(r, "<td align=right>%.02lf</td>\n",
                    h0->tx1debt_max);

                /* Bytes/second. */
                ap_rprintf(r, "<td align=right>%lu</td>\n",
                    (unsigned long)h0->tx2cred_rate);
                /* Bytes. */
                ap_rprintf(r, "<td align=right>%lu</td>\n",
                    (unsigned long)h0->tx2debt_max);
                }

            /* Display internal info about the hash key and table index. */
            if (show_key)
                ap_rprintf(r, "<td align=right>%u</td>", h);
            if (show_index)
                ap_rprintf(r, "<td align=right>%u</td>", i);
            ap_rputs("</tr>\n\n", r);

            /* If doing a slow display, keep pushing bytes out. */
            /* The table doesn't draw faster, but the browser shows progress.*/
            /* On-the-fly DNS lookups may some day have display_rdns == 2. */
            if (display_rdns >= 2)
                ap_rflush(r);
            }
        ap_rputs("</table>\n", r);

        /* Do a quick run through the table to count records. */
        n_host_records = 0;
        for (i = 0, h0 = shm_copy->hosts; i < N_REMOTE_HOSTS; ++i, ++h0) {
            /*-----------------------------------------------------------------
            Note that a single field h0->hash_index indicates whether
            a host record is occupied or not.
            -----------------------------------------------------------------*/
            if (h0->hash_index != NULL_HASH)
                n_host_records += 1;
            }
        n_hash_keys = 0;
        for (i = 0; i < N_HASH_ENTRIES; ++i) {
            if (shm_copy->hashtable[i].head != NULL_HOST)
                n_hash_keys += 1;
            }
        if (show_key_total || show_index_total) {
            if (show_key_total)
                ap_rprintf(r,
                    "Host records = %d/%d.", n_host_records, N_REMOTE_HOSTS);
            if (show_index_total) {
                if (show_key_total)
                    ap_rputs(" ", r);
                ap_rprintf(r,
                    "Hash keys = %d/%d.", n_hash_keys, N_HASH_ENTRIES);
                }
            ap_rputs("<br>\n", r);
            }
        ap_rputs("</center>\n", r);

        /* Flush output, in case the rest of the file transfer is held up. */
        ap_rflush(r);

#if BW_AP_VERSION_1
        ap_unblock_alarms();  /* I don't know if this is needed. */
#endif
        } /* if (shm_base). */
    if (!shm_base) {
        ap_rputs("<p><hr><b>Could not allocate shared memory!</b>\n", r);
        }
    if (trace_level >= 1)
        trace_internal(r);
    ap_rputs("</body></html>\n", r);

    return OK;
    } /* End of function bwshare_trace. */

/*-----------------------------------------------------------------------------
From the point of view of the bwshare module, the ability to serve up
information through file-type handlers like "bwshare_info" is useful only
for management of the module.
Clearly it is straightforward to display stuff to a browser.
It may be a little more difficult to send instructions (in a php-like manner)
from a browser to the module to alter parameters.
Initially, control of the module will be through the httpd.conf file and
other configuration files referred to in httpd.conf.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*     bwshare_info     */
/*----------------------*/
static int bwshare_info(request_rec* r) {
    time_t t1 = 0;
    int trace_level = 0;    /* Trace level for this handler. */

#if BW_AP_VERSION_2
    if (strcmp(r->handler, "bwshare-info") != 0)
        return DECLINED;
#endif

    /* Try to work out how often this is called. */
    n_info += 1;

    /* Dodgy non-protected access to shared memory. */
    if (shm_base) {
        /* Protect access to shared memory with a semaphore. */
        /* +++ Begin Critical Section. +++ */
        sem_grab();
        shm_base->n_info += 1;
        sem_free();
        /* --- End Critical Section. --- */
        }

    /* Stick a header on the output first. */
    r->content_type = "text/html";

    /* Why is the header forced out here? */
#if BW_AP_VERSION_1
    ap_send_http_header(r);
#endif

    /* Is this for the case that "HEAD" is the action requested? */
    if (r->header_only)
        return OK;

    /* Now try to read the command-line arguments for this request. */
    if (r->args && *r->args) {
        char* args = r->args;
        while (args && *args) {
            char* arg1 = 0;
            char* arg2 = 0;

            /* Get the first word of the remaining string. */
            arg1 = ap_getword_nc(r->pool, &args, '&');
            /* ap_getword skips multiple matches. So null-string means end. */
            if (!arg1 || !*arg1)
                break;

            /* Get the string for the variable to set. */
            /* After this call, have arg2 '=' arg1. */
            arg2 = ap_getword_nulls_nc(r->pool, &arg1, '=');

            if ((strcmp(arg2, "trace")) == 0) {
                /* Parameter of form "trace=X". */
                trace_level = atoi(arg1);
                }
            /* Silently ignore any unrecognized parameters. */
            }
        }

    /* Some content. */
    t1 = time((time_t*)0);
    ap_rputs(
        "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
        "<html lang=\"en\"><head>\n"
     "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">\n"
        "<meta name=\"robots\" content=\"noindex,nofollow\">\n"
        "<title>bwshare " BWSHARE_VERSION " info</title>\n"
        "</head>\n"
        "<body bgcolor=\"" COLOUR_INFO_BG "\"><center>\n"
        "<font size=\"+2\"><b>bwshare " BWSHARE_VERSION " info</b><br></font>\n"
        "<a href=\"bwshare-trace\">the bwshare trace page</a><br>\n"
        "<a href=\"http://www.topology.org/src/bwshare/README.html\">"
        "source for this Apache module</a>\n"
        "</center>\n", r);

    if (shm_base) {
        bws_shm_t* shm_copy = 0;
        time_t t0 = 0;
#if BW_AP_VERSION_2
        apr_time_t t0_64;
        apr_time_t t1_64;
#endif

#if BW_AP_VERSION_1
        ap_block_alarms();  /* I don't know if this is needed. */
#endif
        /*---------------------------------------------------------------------
        It seems a little risky to do I/O while lowering the semaphore for
        shared memory. So copy all of it in haste, and print at leisure.
        ---------------------------------------------------------------------*/
#if BW_AP_VERSION_1
        shm_copy = (bws_shm_t*)ap_palloc(r->pool, sizeof(bws_shm_t));
#else
        shm_copy = (bws_shm_t*)apr_palloc(r->pool, sizeof(bws_shm_t));
#endif
        if (!shm_copy) {
            /* This is a bad but unlikely error. */
            ap_rputs(
                "<br>Error: allocation of memory for shm_copy failed!\n"
                "</body></html>\n", r);
#if BW_AP_VERSION_1
            ap_unblock_alarms(); /* Important!! */
#endif
            return OK;
            }

        /* Get a quick copy of the entire shared memory structure. */
        /* +++ Begin Critical Section. +++ */
        sem_grab();
        memcpy(shm_copy, shm_base, sizeof(bws_shm_t));
        sem_free();
        /* --- End Critical Section. --- */

        t0 = shm_copy->start_time;
#if BW_AP_VERSION_2
        apr_time_ansi_put(&t0_64, t0);
        apr_time_ansi_put(&t1_64, t1);
#endif
        ap_rputs("<p><center>"
           "<table rules=all border=1 cellspacing=0 cellpadding=2 bgcolor=\""
            COLOUR_INFO_TAB0 "\">\n", r);

        /* Row 1. */
        ap_rprintf(r,
            "<tr><td>httpd uptime</td><td align=right>%ld sec</td>\n",
            (long)(t1 - t0));
        ap_rprintf(r,
            "<td>&nbsp;<br></td><td>tx1 credit rate<td align=right>"
            "%.02f files/min</td></tr>\n", shm_copy->tx1cred_rate * 60);

        /* Row 2. */
#if BW_AP_VERSION_1
        ap_rprintf(r, "<tr><td>start utc</td><td align=right>%s</td>\n",
            ap_ht_time(r->pool, t0, "%Y-%m-%d %H:%M:%S", 1));
#else
        ap_rprintf(r, "<tr><td>start utc</td><td align=right>%s</td>\n",
            ap_ht_time(r->pool, t0_64, "%Y-%m-%d %H:%M:%S", 1));
#endif
        ap_rprintf(r,
            "<td>&nbsp;<br></td>"
            "<td>tx1 debt max<td align=right>%.02f files</td></tr>\n",
            shm_copy->tx1debt_max);

        /* Row 3. */
#if BW_AP_VERSION_1
        ap_rprintf(r, "<tr><td>current utc</td><td align=right>%s</td>\n",
            ap_ht_time(r->pool, t1, "%Y-%m-%d %H:%M:%S", 1));
#else
        ap_rprintf(r, "<tr><td>current utc</td><td align=right>%s</td>\n",
            ap_ht_time(r->pool, t1_64, "%Y-%m-%d %H:%M:%S", 1));
#endif
        ap_rprintf(r,
            "<td>&nbsp;<br></td><td>tx2 credit rate<td align=right>"
            "%lu bytes/sec</td></tr>\n",
            (unsigned long)shm_copy->tx2cred_rate);

        /* Row 4. */
        ap_rprintf(r, "<tr><td>total requests</td>"
            "<td align=right>%ld files</td>\n", shm_copy->n_logs);
        ap_rprintf(r,
            "<td>&nbsp;<br></td>"
            "<td>tx2 debt max</td><td align=right>%lu bytes</td></tr>\n",
            (unsigned long)shm_copy->tx2debt_max);

        /* Row 5. */
        /* This is a 64-bit integer! Type %qd was introduced in 1.3.7. */
        /* Why does %lld not work? */
        ap_rprintf(r,
            "<tr><td>total data sent</td><td align=right>%s bytes</td>\n",
            uint64comma_string(shm_copy->n_bytes));
        ap_rprintf(r,
            "<td>&nbsp;<br></td>"
            "<td>files warning</td><td align=right>%ld files</td></tr>\n",
            shm_copy->n_files_warn);

        /* Row 6. */
        /* Print the rounded-down bit-rate as integer bits per second. */
        ap_rprintf(r,
            "<tr><td>average bitrate<td align=right>%.0f bits/sec</td>\n",
            (t1 > t0) ? (((double)shm_copy->n_bytes * 8)/(t1 - t0)) : 0);
        ap_rprintf(r,
            "<td>&nbsp;<br><td>bytes warning<td align=right>%ld bytes</td>\n",
            shm_copy->n_bytes_warn);
        ap_rputs("</table>\n\n<p>", r);
#if 0
        /* Probably nobody wants to see the hash table stats on this page. */
        { /* Scoping bracket. */
        int n_host_records = 0;
        int n_hash_keys = 0;
        bws_host_t* h0 = 0;
        int i = 0;

        ap_rputs("<b>Client traffic statistics.</b><br>\n", r);

        /* Do a quick run through the table to count records. */
        n_host_records = 0;
        for (i = 0, h0 = shm_copy->hosts; i < N_REMOTE_HOSTS; ++i, ++h0) {
            /*-----------------------------------------------------------------
            A single field h0->hash_index indicates whether
            a host record is occupied or not.
            -----------------------------------------------------------------*/
            if (h0->hash_index != NULL_HASH)
                n_host_records += 1;
            }
        n_hash_keys = 0;
        for (i = 0; i < N_HASH_ENTRIES; ++i) {
            if (shm_copy->hashtable[i].head != NULL_HOST)
                n_hash_keys += 1;
            }
        ap_rprintf(r, "Host records = %d/%d.\n",
            n_host_records, N_REMOTE_HOSTS);
        ap_rprintf(r, "Hash keys = %d/%d.<br>\n", n_hash_keys, N_HASH_ENTRIES);
        } /* Scoping bracket. */
#endif
        /* Put up a form for the operator to choose from. */
        /* Note: Should put this in a function. */
        ap_rputs("<form name=\"bwshare1\" method=\"GET\" "
            "action=\"bwshare-trace\">\n"
            "<table rules=all border=1 cellspacing=0 cellpadding=6 bgcolor=\""
            COLOUR_INFO_FRM_BD "\">"
            "<tr valign=top><td>\n", r);

        ap_rputs("<table cellspacing=3 cellpadding=2 border=0 bgcolor=\""
            COLOUR_INFO_FRM_BG "\">\n"
            "<tr><td align=center colspan=2>"
            "<b><i>parameters for trace screen</i></b></td></tr>\n", r);

        /* Row 2. Column 1. */
        ap_rputs("<tr><td valign=middle>sort field<br>\n"
            "<input name=sort type=radio value=\"ip\"> ip<br>\n"
            "<input name=sort type=radio value=\"files\"> files<br>\n"
            "<input name=sort type=radio value=\"bytes\"> bytes<br>\n"
          "<input name=sort type=radio value=\"idle\" checked> idle time<br>\n"
            "<input name=sort type=radio value=\"tx1\"> TX 1 debt<br>\n"
          "<input name=sort type=radio value=\"tx1peak\"> TX 1 debt peak<br>\n"
            "<input name=sort type=radio value=\"tx2\"> TX 2 debt<br>\n"
          "<input name=sort type=radio value=\"tx2peak\"> TX 2 debt peak<br>\n"
            "<input name=sort type=radio value=\"rdns\">"
            " <i style=\"color:" COLOUR_INFO_FRM_FADE ";\">reverse DNS</i>"
            "<br>\n"
            "<input name=sort type=radio value=\"key\"> hash key<br>\n"
            "<input name=sort type=radio value=\"index\"> index<br>\n"
            "</td>\n", r);

        /* Row 2. Column 2. */
        ap_rputs(
            "<td valign=middle align=center><table border=0 height=\"100%\">",
            r);

        ap_rputs(
            "<tr><td align=right>\n"
       "show reverse DNS <input name=rdns type=radio value=\"1\" checked><br>\n"
            "hide reverse DNS <input name=rdns type=radio value=\"0\"><br>\n"
            "<br>\n"
            "short list <input name=set type=radio value=\"1\" checked><br>\n"
            "full list <input name=set type=radio value=\"0\"><br>\n"
            "<br>\n"
            "show bw limits <input name=limits type=radio value=\"1\"><br>\n"
       "hide bw limits <input name=limits type=radio value=\"0\" checked><br>\n"
            "<br>\n"
            "</td></tr>\n", r);

        ap_rprintf(r,
            "<tr><td align=left>"
            "<input name=\"recent\" type=text size=\"6\" value=\"%lu\">"
            "&nbsp;&nbsp;recent time (seconds)"
            "</td></tr>\n",
            (unsigned long)shm_copy->idle_show);

        ap_rputs(
            "<tr><td align=left>"
            "<input name=\"trace\" type=text size=\"6\" value=\"0\">"
            "&nbsp;&nbsp;trace level"
            "</td></tr>\n", r);

        ap_rputs("</table></td></tr>\n", r);

        /* Row 3. */
        ap_rputs(
            "<tr valign=bottom><td align=left>\n"
            "<input name=\"reset\" type=\"reset\" value=\"reset\"></td>\n"
            "<td align=right valign=bottom>"
            "<input name=\"b\" type=\"submit\" value=\"trace\">"
            "</td></tr>\n" , r);
        ap_rputs("</table>\n", r);
        ap_rputs("</td></tr></table></form></center>\n", r);

        /* Flush output, in case the rest of the file transfer is held up. */
        ap_rflush(r);
#if BW_AP_VERSION_1
        ap_unblock_alarms();  /* I don't know if this is needed. */
#endif
        }
    if (!shm_base) {
        ap_rputs("\n<p><hr><b>Could not allocate shared memory!</b>\n", r);
        }
    if (trace_level >= 1)
        trace_internal(r);
    ap_rputs("</body></html>\n", r);

    return OK;
    } /* End of function bwshare_info. */

/*-----------------------------------------------------------------------------
bwshare_access_check() determines if the user is over budget.
If so, the file download is refused.
The status codes are defined in the HTTP/1.1 spec, RFC 2616.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Note that if an error condition is to be returned, it would be preferable to
return a message specific to "bwshare", so that the user will know what is
going on. A little bit of quantitative information could be returned even.
But not _too_ much!
-----------------------------------------------------------------------------*/
/*----------------------*/
/* bwshare_access_check */
/*----------------------*/
static int bwshare_access_check(request_rec* r) {
#if BW_AP_VERSION_2
    struct sockaddr_in* sinp = 0; /* Internet socket address. */
    apr_sockaddr_t* sockp = 0;
#endif
    uint32 rem_addr = 0;
    time_t t1 = 0;
    int host_index = NULL_HOST;
    int hash_index = NULL_HASH;
    int ret_value = OK;
    double delay_value = 0;
    int delay_reason = 0;
    bws_context_t* context_dir = 0; /* Directory context. */
    bws_context_t* context_srv = 0; /* Virtual host context. */
    int throttle_off = 0;

    /* Check sanity of arguments. */
    if (!r || !r->connection || !shm_base)
        return DECLINED;    /* Probably will never happen. */

    /* Get the per-directory configuration settings from request_rec. */
    /* Note: Not currently used. */
    if (r->per_dir_config) {
        context_dir = (bws_context_t*)
            ap_get_module_config(r->per_dir_config, &bwshare_module);
        }
    /* Get the per-virtual-server configuration settings from server_rec. */
    if (r->server && r->server->module_config) {
        context_srv = (bws_context_t*)
            ap_get_module_config(r->server->module_config, &bwshare_module);
        }

    /* Have a look at the virtual host config for non-standard settings. */
    if (context_srv)
        throttle_off = context_srv->throttle_off;

    /* Note: If throttle_off is set, then the rest of this function could */
    /* be skipped because there will be no side-effects. */

    /* Determine the remote IP address. */
#if BW_AP_VERSION_1
    rem_addr = ntohl(r->connection->remote_addr.sin_addr.s_addr);
#else
    /* remote_addr is of type apr_sockaddr_t. */
    /* This type is defined in srclib/apr/include/apr_network_io.h. */
    if (!r->connection->remote_addr)
        return DECLINED;
    sockp = r->connection->remote_addr;
    sinp = &sockp->sa.sin;

#if BW_CONN_TRACE
    /* Some some ad-hoc tracing. */
#if APR_HAVE_IPV6
    ctt_handler.family = sockp->sa.sin6.sin6_family;
#else
    ctt_handler.family = 0;
#endif
    ctt_handler.port = sockp->port;
    ctt_handler.salen = sockp->salen;
    ctt_handler.ipaddr_len = sockp->ipaddr_len;
    ctt_handler.addr_str_len = sockp->addr_str_len;
#endif
    if (sinp->sin_family == AF_INET) {
        rem_addr = ntohl(sinp->sin_addr.s_addr);
        }
    else if (sinp->sin_family == AF_INET6) {
        /* IPv6 not yet supported. So all IPv6 gets through scot-free!! */
        /* Temporarily ignore non-IPv4 addresses. */
/*
        if (!IN6_IS_ADDR_V4MAPPED((struct in6_addr*)sockp->ipaddr_ptr))
            return DECLINED;
*/
        rem_addr = ntohl(((uint32_t*)sockp->ipaddr_ptr)[3]);
        }
#endif

    /* Calculate the hash key for this remote IP address. */
    hash_index = crc16calc_rl(rem_addr) % N_HASH_ENTRIES;
    if (hash_index >= N_HASH_ENTRIES)
        hash_index = 0;     /* Should never happen. */

    /* Get the current time since UTC 1970 January 1. */
    t1 = time((time_t*)0);

    /* Must not return or stall between here and the sem_free() call. */
    /* +++ Begin Critical Section. +++ */
    sem_grab();

    /* Get the host index for head-of-list from the hash index. */
    host_index = shm_base->hashtable[hash_index].head;

    /* If a host record list is found, seek the right host record. */
    /* This should be in a separate function maybe. */
    if (host_index != NULL_HOST) {
        /* Follow the list to the right record. */
        while (shm_base->hosts[host_index].address != rem_addr) {
            /* Assume that there is no corruption of next-pointers. */
            /* In particular, assume that "next" is always within range. */
            host_index = shm_base->hosts[host_index].next;

            /* If the list exhausts, it just isn't there. */
            if (host_index == NULL_HOST)
                break;
            }
        /* If no record was found in this list, don't create a new record. */
        }

    /* If the remote host has a record, scrutinize it for debt. */
    if (host_index != NULL_HOST) {
        double tx1debt;
        double tx2debt;
        double tx1debt_max = shm_base->hosts[host_index].tx1debt_max;
        double tx2debt_max = shm_base->hosts[host_index].tx2debt_max;
        double dt = t1 - shm_base->hosts[host_index].last_time;

        /* Evaluate current TX debt value tx1debt, but don't update it. */
        /* Reduce debt for each second elapsed. */
        tx1debt = shm_base->hosts[host_index].tx1debt
                - shm_base->hosts[host_index].tx1cred_rate * dt;
        if (tx1debt < 0)
            tx1debt = 0;

        /* Evaluate current TX debt value tx2debt, but don't update it. */
        /* Reduce debt for each second elapsed. */
        tx2debt = shm_base->hosts[host_index].tx2debt
                - shm_base->hosts[host_index].tx2cred_rate * dt;
        if (tx2debt < 0)
            tx2debt = 0;

        if (!throttle_off &&
                ((tx1debt > tx1debt_max) || (tx2debt > tx2debt_max))) {
            /* Calculate the time to recovery. */
            /* Adjust by 1 file because of the current request. */
            double delay1 =
                (tx1debt + 1 - shm_base->hosts[host_index].tx1debt_max)
                / shm_base->hosts[host_index].tx1cred_rate;
            double delay2 =
                (tx2debt - shm_base->hosts[host_index].tx2debt_max)
                / shm_base->hosts[host_index].tx2cred_rate;
            if (delay1 > delay2) {
                delay_value = delay1;
                delay_reason = 1;
                }
            else {
                delay_value = delay2;
                delay_reason = 2;
                }

            /* Muaaaahahahahahaha!!! (Rough quote from Anthony Howe, 1999.) */
            ret_value = HTTP_SERVICE_UNAVAILABLE;
            }
        }
    sem_free();
    /* --- End Critical Section. --- */

    /* Replace the standard message with my own message. */
    if (ret_value == HTTP_SERVICE_UNAVAILABLE) {
        /* The 503 causes some rampant downloaders to try forever. */
        r->status = 503;
/*        r->status = 200; */
        /*---------------------------------------------------------------------
        If you put a \n on the end of the status line, the remaining headers
        of the server response will go in the document body!! So don't do it.
        ---------------------------------------------------------------------*/
        r->status_line = "503 Service Temporarily Unavailable - Overload";
/*        r->status_line = "200 Service Temporarily Unavailable - Overload"; */
        r->content_type = "text/html";
        ap_discard_request_body(r);     /* Huh? */
#if BW_AP_VERSION_1
        ap_send_http_header(r);
#endif

        /* Is this for the case that "HEAD" is the action requested? */
        if (! r->header_only) {
            /* Basic fixed header. */
            ap_rputs(
          "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
                "<html lang=\"en\"><head>\n"
              "<meta name=\"robots\" content=\"noindex,nofollow,noarchive\">\n"
                "<title>503 Service Temporarily Unavailable</title>\n"
                "</head>\n<body>\n"
                "<h1>Service Temporarily Unavailable</h1>\n", r);

            /* The real message for the user. */
            ap_rprintf(r,
                "The bwshare module will refuse your requests"
                " for the next %.0f seconds.<br>\n", delay_value);
            if (delay_reason == 1) {
                ap_rputs(
                    "You have made too many requests per second.\n", r);
                }
            else if (delay_reason == 2)
                ap_rputs("You have downloaded data too rapidly.\n", r);

            /* Basic trailer. */
            ap_rputs(ap_psignature("\n<p><hr>\n", r), r);
            ap_rputs("</body></html>\n", r);
            }
        ret_value = DONE;               /* What??! */
        }
    return ret_value;
    } /* End of function bwshare_access_check. */

/*-----------------------------------------------------------------------------
bwshare_logger() is supposed to be called after a file has been completely
served to the client.
This implies that if a client downloads a very big file, the amount downloaded
won't be recorded for a long time.
So the client could initiate many very large file transfers at the same time
without having their downloads recorded.
Maybe it is not knowable in advance how many bytes are expected to be in any
particular file request.
Should try to determine the number of bytes in a download _before_ the
download, so that it can be refused or delayed before it gets started.
-----------------------------------------------------------------------------*/
/*----------------------*/
/*     bwshare_logger   */
/*----------------------*/
static int bwshare_logger(request_rec* r) {
    conn_rec* conn = 0;
#if BW_AP_VERSION_2
    struct sockaddr_in* sinp = 0; /* Internet socket address. */
    apr_sockaddr_t* sockp = 0;
#endif
    uint32 rem_addr = 0;
    int host_index = NULL_HOST;
    int hash_index = NULL_HASH;
    time_t t1 = 0;          /* The current time. */
    const char* rem_name = 0;  /* Reverse DNS lookup. */

#if 1
    uint64 txbytes = 0; /* Number of bytes really sent to client. */
#else
    /* apr_off_t is defined in "srclib/apr/include/apr.h.in". */
    /* Apparently not available in Apache 1. */
    apr_off_t txbytes = 0; /* Number of bytes really sent to client. */
#endif

    /* Sanity check. */
    if (!r || !shm_base)
        return DECLINED;

    /* "clength" might be right here? Someone should document the Apache API! */
    if (r && r->sent_bodyct) {
#if BYTES_SENT
        txbytes = (uint64)r->bytes_sent;
#else
        txbytes = (uint64)r->read_length;
#endif
        }

    /*-------------------------------------------------------------------------
    Update some aggregate stats.
    If we're interrupted between here and the host update, the aggregate
    stats won't agree with the sum of host stats. But this is not important.
    -------------------------------------------------------------------------*/
    /* +++ Begin Critical Section. +++ */
    sem_grab();
    shm_base->n_logs += 1;
    shm_base->n_bytes += txbytes;
    sem_free();
    /* --- End Critical Section. --- */

    /* Record the number of bytes the client obtained. */
    if (!r)
        return DECLINED;    /* Probably will never happen. */
    conn = r->connection;
    if (!conn)
        return DECLINED;    /* Probably this won't ever happen either. */
#if BW_AP_VERSION_1
    rem_addr = ntohl(conn->remote_addr.sin_addr.s_addr);
#else
    /* remote_addr is of type apr_sockaddr_t. */
    /* This type is defined in srclib/apr/include/apr_network_io.h. */
    if (!conn->remote_addr)
        return DECLINED;
    sockp = conn->remote_addr;
    sinp = &sockp->sa.sin;

#if BW_CONN_TRACE
    /* Some some ad-hoc tracing. */
#if APR_HAVE_IPV6
    ctt_logger.family = sockp->sa.sin6.sin6_family;
#else
    ctt_logger.family = 0;
#endif
    ctt_logger.port = sockp->port;
    ctt_logger.salen = sockp->salen;
    ctt_logger.ipaddr_len = sockp->ipaddr_len;
    ctt_logger.addr_str_len = sockp->addr_str_len;
#endif
    if (sinp->sin_family == AF_INET) {
        /* IPv6 not yet supported. So all IPv6 gets through scot-free!! */
        rem_addr = ntohl(sinp->sin_addr.s_addr);
        }
    else if (sinp->sin_family == AF_INET6) {
        /* sockaddr_in6 and sockaddr_in are defined in netinet.in.h. */
        /* Temporarily ignore non-IPv4 addresses. */
/*
        if (!IN6_IS_ADDR_V4MAPPED((struct in6_addr*)sockp->ipaddr_ptr))
            return DECLINED;
*/
        /* If it's not V4 mapped, just use the last 32 bits anyway! */
        /* Note: I should convert this software to 128-bit addresses!! */
        rem_addr = ntohl(((uint32_t*)sockp->ipaddr_ptr)[3]);
        }
    else {
#if APR_HAVE_IPV6
        rem_addr = ntohl(sockp->sa.sin6.sin6_addr.s6_addr32[3]);
#else
        rem_addr = ntohl(((uint32_t*)sockp->ipaddr_ptr)[3]);
#endif
        }
#endif

    /*-------------------------------------------------------------------------
    Get reverse DNS translations from the server.
    Should somehow arrange to only do this if the DNS is not yet cached.
    Hopefully it has already been obtained by the server.
    -------------------------------------------------------------------------*/
#if BW_AP_VERSION_1
    rem_name = ap_get_remote_host(conn, (void*)0, REMOTE_HOST);
#else
    rem_name = ap_get_remote_host(conn, (void*)0, REMOTE_HOST, (int*)0);
#endif

    /* Calculate the appropriate hash key for this IP address. */
    hash_index = crc16calc_rl(rem_addr) % N_HASH_ENTRIES;
    if (hash_index >= N_HASH_ENTRIES)
        hash_index = 0;     /* Should never happen. */

    /* Get the current time since UTC 1970 January 1. */
    t1 = time((time_t*)0);

    /* Must not return or stall between here and the sem_free() call. */
    /* +++ Begin Critical Section. +++ */
    sem_grab();

    /* Get the host index from the hash index. */
    host_index = shm_base->hashtable[hash_index].head;

    /* Try to find or create a host record for this host. */
    if (host_index == NULL_HOST) {
        /* Case that the list for this hash is empty. */
        /* Try to allocate a new host record. */
        host_index = bws_newhost(shm_base);

        /* If successful, set some relevant fields in the record. */
        if (host_index != NULL_HOST) {
            bws_host_t* h0 = &shm_base->hosts[host_index];
            int i = 0;

            /* Record the new host record in the hash table. */
            shm_base->hashtable[hash_index].head = host_index;

            /* Set some attributes. */
            h0->hash_index = hash_index;
            h0->address = rem_addr;
            h0->last_time = t1;
            host_rdns_set(h0, rem_name);

            /* See if there are special bw params for this host. */
            /* NOTE: This should be a function. */
            for (i = 0; i < shm_base->n_networks; ++i) {
                bws_net_t* subnet = &shm_base->networks[i];
                /* Use the first match in the list. */
                if (((rem_addr ^ subnet->address) & subnet->mask) == 0) {
                    h0->tx1cred_rate = subnet->tx1cred_rate;
                    h0->tx1debt_max = subnet->tx1debt_max;
                    h0->tx2cred_rate = subnet->tx2cred_rate;
                    h0->tx2debt_max = subnet->tx2debt_max;
                    break;
                    }
                }
            }
        }
    else {
        /* If an _old_ host record _list_ has been found, seek the host. */
        /* Follow the list to the right record. */
        while (shm_base->hosts[host_index].address != rem_addr) {
            /* Assume that there is no corruption of next-pointers. */
            /* In particular, assume that "next" is always within range. */
            host_index = shm_base->hosts[host_index].next;

            /* If the list exhausts, must allocate a new host record. */
            if (host_index == NULL_HOST)
                break;
            }
        /* If no record was found in this list, try to create a new record. */
        if (host_index == NULL_HOST) {
            /* Try to allocate a new host record. */
            host_index = bws_newhost(shm_base);

            /* If successful, set some relevant fields in the record. */
            if (host_index != NULL_HOST) {
                bws_host_t* h0 = &shm_base->hosts[host_index];
                int i = 0;

                /* Must link this new host record into the per-key list. */
                /* Put the new record at the head of the list. */
                h0->next = shm_base->hashtable[hash_index].head;

                /* Record the new host record in the hash table. */
                shm_base->hashtable[hash_index].head = host_index;

                /* Set some attributes. */
                h0->hash_index = hash_index;
                h0->address = rem_addr;
                h0->last_time = t1;
                host_rdns_set(h0, rem_name);

                for (i = 0; i < shm_base->n_networks; ++i) {
                    bws_net_t* subnet = &shm_base->networks[i];
                    /* Use the first match in the list. */
                    if (((rem_addr ^ subnet->address) & subnet->mask) == 0) {
                        h0->tx1cred_rate = subnet->tx1cred_rate;
                        h0->tx1debt_max = subnet->tx1debt_max;
                        h0->tx2cred_rate = subnet->tx2cred_rate;
                        h0->tx2debt_max = subnet->tx2debt_max;
                        break;
                        }
                    }
                }
            }
        }

    /* If the host record _has_ been found or created, update it. */
    if (host_index != NULL_HOST) {
        double tx1debt;
        double tx2debt;
        time_t t0 = shm_base->hosts[host_index].last_time;

/*        shm_base->hosts[host_index].address = rem_addr; */
        shm_base->hosts[host_index].n_files += 1;
        if (r->sent_bodyct)
            shm_base->hosts[host_index].n_bytes += txbytes;
        /* Update the TX debt value tx1debt. Uses last_time. */
        /* Reduce debt for each second elapsed. */
        tx1debt = shm_base->hosts[host_index].tx1debt
                - shm_base->hosts[host_index].tx1cred_rate * (t1 - t0);
        if (tx1debt < 0)
            tx1debt = 0;

        /* Increase debt because of file transmission. */
        tx1debt += 1;
        shm_base->hosts[host_index].tx1debt = tx1debt;
        if (shm_base->hosts[host_index].tx1debtpeak < tx1debt)
            shm_base->hosts[host_index].tx1debtpeak = tx1debt;

        /* Update the TX debt value tx2debt. Uses last_time. */
        /* Reduce debt for each second elapsed. */
        tx2debt = shm_base->hosts[host_index].tx2debt
                - shm_base->hosts[host_index].tx2cred_rate * (t1 - t0);
        if (tx2debt < 0)
            tx2debt = 0;

        /* Increase debt because of file transmission. */
        if (r->sent_bodyct)
            tx2debt += txbytes;
        shm_base->hosts[host_index].tx2debt = tx2debt;
        if (shm_base->hosts[host_index].tx2debtpeak < tx2debt)
            shm_base->hosts[host_index].tx2debtpeak = tx2debt;

        /* Use the current time rather than the request time. */
        /* This is because the request time might not increase monotonically.*/
/*        shm_base->hosts[host_index].last_time = r->request_time; */
        shm_base->hosts[host_index].last_time = t1;
        }
    sem_free();
    /* --- End Critical Section. --- */

    /* Indicate that there is no objection to the request continuing. */
    return DECLINED;
    } /* End of function bwshare_logger. */

#if BW_AP_VERSION_2
/*--------------------------*/
/*   bwshare_init_handler   */
/*--------------------------*/
static int bwshare_init_handler(apr_pool_t* p, apr_pool_t* plog,
                                apr_pool_t* ptemp, server_rec* s) {
    ap_add_version_component(p, "mod_bwshare" "/" BWSHARE_VERSION);
#if ! BW_INIT_NEW_LOCATION
    /* This is moved to bwshare_create_server() for no obvious reason. */
    bwshare_init(p, plog, ptemp, s);
#endif
    return OK;
    } /* End of function bwshare_init_handler. */
#endif /* BW_AP_VERSION_2 */

#if BW_AP_VERSION_2
/*--------------------------*/
/*  bwshare_register_hooks  */
/*--------------------------*/
static void bwshare_register_hooks(apr_pool_t* p) {
/*
    static const char* const aszPre[] = { "http_core.c", NULL };
*/
    /* The following line had bwshare_init for Apache version 2.0.16. */
    ap_hook_post_config(bwshare_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_access_checker(bwshare_access_check, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_log_transaction(bwshare_logger, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_handler(bwshare_trace, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_handler(bwshare_info, NULL, NULL, APR_HOOK_MIDDLE);
    }; /* End of function bwshare_register_hooks. */
#endif /* BW_AP_VERSION_2 */

/*-----------------------------------------------------------------------------
The httpd.conf file should include something like this:

<Location /bwshare-info>
    SetHandler bwshare-info
</Location>

<Location /bwshare-trace>
    SetHandler bwshare-trace
</Location>

This means that requests for the file /bwshare-info will be handled
by the handler function "bwshare-info", which is mapped in the handler_rec
array to the C function "bwshare_info".
Similarly, I could do

<Directory />
    AddHandler bwshare-info bws
</Directory>

to cause the bwshare_info function to be invoked for all files ending
with the extension ".bws". (This is not something which I would do.)
-----------------------------------------------------------------------------*/

#if BW_AP_VERSION_1
static const handler_rec bwshare_handlers[] = {
    { "bwshare-trace",      bwshare_trace },
    { "bwshare-info",       bwshare_info },
    { NULL }
    }; /* End of struct handler_rec bwshare_handlers[]. */
#endif /* BW_AP_VERSION_1 */

#if BW_AP_VERSION_1
#define CR_L
#define CR_R
#define CR_C
#endif /* BW_AP_VERSION_1 */

#if BW_AP_VERSION_2
/* Apache 2 defines the handler as "union cmd_func" in http_config.h. */
#define CR_L {
#define CR_R }
/* To initialize a union, you have to use the type of first element. */
#define CR_C (const char* (*)(cmd_parms*, void*))
#endif /* BW_AP_VERSION_2 */

/*---------------------------------------------------------------------------
For definition of TAKE1, RAW_ARGS etc., see:
http://www.aprapi.com/http__config_8h.html

For OR_FILEINFO etc., see:
http://www.aprapi.com/group__ConfigDirectives.html

The info strings in bwshare_cmds are displayed by mod_info.c.
---------------------------------------------------------------------------*/
static const command_rec bwshare_cmds[] = {
    { "BW_tx1debt_max", CR_L CR_C bw_tx1debt_max CR_R, NULL,
      OR_FILEINFO, TAKE1, "the max number of files debt (TX1)" },
    { "BW_tx1cred_rate", CR_L CR_C bw_tx1cred_rate CR_R, NULL,
      OR_FILEINFO, TAKE1, "the max file service rate (TX1)" },
    { "BW_tx2debt_max", CR_L CR_C bw_tx2debt_max CR_R, NULL,
      OR_FILEINFO, TAKE1, "the max number of bytes debt (TX2)" },
    { "BW_tx2cred_rate", CR_L CR_C bw_tx2cred_rate CR_R, NULL,
      OR_FILEINFO, TAKE1, "the max file service rate (TX2)" },
    { "BW_throttle_off", CR_L CR_C bw_throttle_off CR_R, NULL,
      OR_FILEINFO, TAKE1, "turn off throttling" },
    { "BW_subnet_limit", CR_L CR_C bw_subnet_limit CR_R, NULL,
      OR_FILEINFO, RAW_ARGS, "bandwidth limitation for client IP subnet" },
    { NULL }
    }; /* End of struct command_rec bwshare_cmds[]. */

/* Type-unsafe kludges! But helps things stay tidy and readable. */
/* These are the module entry points and/or tables that are not yet defined. */
#define bwshare_create_dir          NULL
#define bwshare_merge_dir           NULL

#define bwshare_translate_handler   NULL
#define bwshare_check_user_id       NULL
#define bwshare_auth_check          NULL

#define bwshare_type_check          NULL
#define bwshare_fixer_upper         NULL

#define bwshare_header_parser       NULL
#define bwshare_child_init          NULL
#define bwshare_child_exit          NULL
#define bwshare_post_read_request   NULL

/*-----------------------------------------------------------------------------
Standard module definition.
This bit is directly adapted from the Apache demo module called "example".
The definitive definition is in http_config.h.
-----------------------------------------------------------------------------*/
#if BW_AP_VERSION_1
/*--------------------------------------------*/
/* This is the version for Apache 1.2.13 etc. */
/*--------------------------------------------*/
module MODULE_VAR_EXPORT bwshare_module = {
    STANDARD_MODULE_STUFF,
    bwshare_init,                   /* module initializer */
    bwshare_create_dir,             /* per-directory config creator */
    bwshare_merge_dir,              /* dir config merger */
    bwshare_create_server,          /* server config creator */
    bwshare_merge_server,           /* server config merger */
    bwshare_cmds,                   /* command table */
    bwshare_handlers,               /* [7] list of handlers */
    bwshare_translate_handler,      /* [2] filename-to-URI translation */
    bwshare_check_user_id,          /* [5] check/validate user_id */
    bwshare_auth_check,             /* [6] check user_id is valid *here* */
    bwshare_access_check,           /* [4] check access by host address */
    bwshare_type_check,             /* [7] MIME type checker/setter */
    bwshare_fixer_upper,            /* [8] fixups */
    bwshare_logger,                 /* [10] logger */
#if MODULE_MAGIC_NUMBER >= 19970103
    bwshare_header_parser,          /* [3] header parser */
#endif
#if MODULE_MAGIC_NUMBER >= 19970719
    bwshare_child_init,             /* process initializer */
#endif
#if MODULE_MAGIC_NUMBER >= 19970728
    bwshare_child_exit,             /* process exit/cleanup */
#endif
#if MODULE_MAGIC_NUMBER >= 19970902
    bwshare_post_read_request       /* [1] post read_request handling */
#endif
    }; /* End of struct module bwshare_module. */
#endif /* BW_AP_VERSION_1 */

#if BW_AP_VERSION_2
/*---------------------------------------------------------------------------
These are the fields in the "module" structure.

  void *(*create_dir_config) (apr_pool_t *p, char *dir);
  void *(*merge_dir_config) (apr_pool_t *p, void *base_conf, void *new_conf);
  void *(*create_server_config) (apr_pool_t *p, server_rec *s);
  void *(*merge_server_config) (apr_pool_t *p, void *base_conf, void *new_conf);
  const command_rec *cmds;
  void (*register_hooks) (apr_pool_t *p);
---------------------------------------------------------------------------*/
/*--------------------------------------------*/
/* This is the version for Apache 2.0.16 etc. */
/*--------------------------------------------*/
/* struct module is defined in http_config.h. */
module AP_MODULE_DECLARE_DATA bwshare_module = {
    STANDARD20_MODULE_STUFF,        /* maj/min,-1,file,0,0,cookie,null. */

    bwshare_create_dir,             /* per-directory config creator */
    bwshare_merge_dir,              /* dir config merger */
    bwshare_create_server,          /* server config creator */
    bwshare_merge_server,           /* server config merger */
    bwshare_cmds,                   /* command table */
    bwshare_register_hooks          /* register-hooks function */
    }; /* End of struct module bwshare_module. */
#endif /* BW_AP_VERSION_2 */
